"""
Line-based and matrix-based plotting commands using MatPlotLib.

Before importing this file, you will probably want to do something
like:

  from matplotlib import rcParams
  rcParams['backend']='TkAgg'

to select a backend, or else select an appropriate one in your
matplotlib.rc file (if any).  There are many backends available for
different GUI or non-GUI uses.

$Id$
"""

import param
    
try:
    import matplotlib.ticker
    import pylab
except ImportError:
    param.Parameterized(name=__name__).warning("Could not import matplotlib; module will not be useable.")
    from topo.command import ImportErrorRaisingFakeModule
    pylab = ImportErrorRaisingFakeModule("matplotlib")  # pyflakes:ignore (try/except import)


import numpy
from math import pi
# JABALERT: Import all of these from numpy instead?
from numpy.oldnumeric import sqrt, array, transpose, argmin, cos, sin, log10, Float
from numpy import outer,arange,ones,zeros

from numpy.fft.fftpack import fft2
from numpy.fft.helper import fftshift
from numpy import abs

import topo
from topo.base.sheetview import SheetView
from topo.base.arrayutil import centroid, wrap
from topo.base.sheet import Sheet
from topo.misc.util import frange
import topo.analysis.vision
from topo.plotting.plot import make_template_plot
import param
from param import ParameterizedFunction,normalize_path
from param.parameterized import ParamOverrides
from topo.pattern import SineGrating, OrientationContrast
from topo.plotting.plotgroup import create_plotgroup
from topo.base.cf import CFSheet

from topo.analysis.featureresponses import Feature, PatternPresenter
from topo.analysis.featureresponses import PositionMeasurementCommand, FeatureCurveCommand, UnitCurveCommand


from topo.command import Command





class PylabPlotCommand(Command):
    """Parameterized command for plotting using Matplotlib/Pylab."""
      
    file_dpi = param.Number(
        default=100.0,bounds=(0,None),softbounds=(0,1000),doc="""
        Default DPI when rendering to a bitmap.  
        The nominal size * the dpi gives the final image size in pixels.  
        E.g.: 4"x4" image * 80 dpi ==> 320x320 pixel image.""")
                                      
    file_format = param.String(default="png",doc="""
        Which image format to use when saving images.
        The output can be png, ps, pdf, svg, or any other format
        supported by Matplotlib.""")

    # JABALERT: Should replace this with a filename_format and
    # associated parameters, as in PlotGroupSaver.  
    # Also should probably allow interactive display to be controlled
    # separately from the filename, to make things work more similarly
    # with and without a GUI.
    filename = param.String(default=None,doc="""
        Optional base of the filename to use when saving images;
        if None the plot will be displayed interactively.
        
        The actual name is constructed from the filename base plus the
        suffix plus the current simulator time plus the file_format.""")

    filename_suffix = param.String(default="",doc="""
        Optional suffix to be used for disambiguation of the filename.""")

    title = param.String(default=None,doc="""
        Optional title to be used when displaying the plot interactively.""")

    __abstract = True


    def _set_windowtitle(self,title):
        """
        Helper function to set the title (if not None) of this PyLab plot window.
        """
        
        # At the moment, PyLab does not offer a window-manager-independent
        # means for controlling the window title, so what we do is to try
        # what should work with Tkinter, and then suppress all errors.  That
        # way we should be ok when rendering to a file-based backend, but
        # will get nice titles in Tk windows.  If other toolkits are in use,
        # the title can be set here using a similar try/except mechanism, or
        # else there can be a switch based on the backend type.
        if title is not None:
            try: 
                manager = pylab.get_current_fig_manager()
                manager.window.title(title)
            except:
                pass
    
    
    def _generate_figure(self,p):
        """
        Helper function to display a figure on screen or save to a file.
        
        p should be a ParamOverrides instance containing the current
        set of parameters.
        """

        pylab.show._needmain=False
        if p.filename is not None:
            # JABALERT: need to reformat this as for other plots
            fullname=p.filename+p.filename_suffix+str(topo.sim.time())+"."+p.file_format
            pylab.savefig(normalize_path(fullname), dpi=p.file_dpi)
        else:
            self._set_windowtitle(p.title)
            pylab.show()



class vectorplot(PylabPlotCommand):
    """
    Simple line plotting for any vector or list of numbers.

    Intended for interactive debugging or analyzing from the command
    prompt.  See MatPlotLib's pylab functions to create more elaborate
    or customized plots; this is just a simple example.

    An optional string can be supplied as a title for the figure, if
    desired.  At present, this is only used for the window, not the
    actual body of the figure (and will thus not appear when the
    figure is saved).

    The style argument allows different line/linespoints style for
    the plot: 'r-' for red solid line, 'bx' for blue x-marks, etc.
    See http://matplotlib.sourceforge.net/matplotlib.pylab.html#-plot
    for more possibilities.

    The label argument can be used to identify the line in a figure legend.

    Ordinarily, the x value for each point on the line is the index of
    that point in the vec array, but a explicit list of xvalues can be
    supplied; it should be the same length as vec.

    Execution of multiple vectorplot() commands with different styles
    will result in all those styles overlaid on a single plot window.
    """

    # JABALERT: All but the first two arguments should probably be Parameters
    def __call__(self,vec,xvalues=None,style='-',label=None,**params):
        p=ParamOverrides(self,params)

        if xvalues is not None:
            pylab.plot(xvalues, vec, style, label=label)
        else:
            pylab.plot(vec, style, label=label)
            
        pylab.grid(True)
        self._generate_figure(p)



class matrixplot(PylabPlotCommand):
    """
    Simple plotting for any matrix as a bitmap with axes.

    Like MatLab's imagesc, scales the values to fit in the range 0 to 1.0.
    Intended for interactive debugging or analyzing from the command
    prompt.  See MatPlotLib's pylab functions to create more elaborate
    or customized plots; this is just a simple example.
    """

    plot_type = param.Callable(default=pylab.gray,doc="""
        Matplotlib command to generate the plot, e.g. pylab.gray or pylab.hsv.""")
    
    extent = param.Parameter(default=None,doc="""
        Subregion of the matrix to plot, as a tuple (l,b,r,t).""")
    
    # JABALERT: All but the first two should probably be Parameters
    def __call__(self,mat,aspect=None,colorbar=True,**params):
        p=ParamOverrides(self,params)
        
        p.plot_type()
        pylab.figure(figsize=(5,5))

        # Swap lbrt to lrbt to match pylab
        if p.extent is None:
            extent = None
        else:
            (l,b,r,t)=p.extent
            extent=(l,r,b,t)
            
        pylab.imshow(mat,interpolation='nearest',aspect=aspect,extent=extent)
        if colorbar and (mat.min()!= mat.max()): pylab.colorbar()
        self._generate_figure(p)


class matrixplot3d(PylabPlotCommand):
    """
    Simple plotting for any matrix as a 3D wireframe with axes.

    Uses Matplotlib's beta-quality features for 3D plotting.  These
    usually work fine for wireframe plots, although they don't always
    format the axis labels properly, and do not support removal of
    hidden lines.  Note that often the plot can be rotated within the
    window to make such problems go away, and then the best result can
    be saved if needed.

    Other than the default "wireframe", the type can be "contour" to
    get a contour plot, or "surface" to get a solid surface plot, but
    surface plots currently fail in many cases, e.g. for small
    matrices.

    If you have trouble, you can try matrixplot3d_gnuplot instead.
    """
    
    # JABALERT: All but the first two arguments should probably be Parameters
    def __call__(self,mat,type="wireframe",**params):
        p=ParamOverrides(self,params)
    
        from mpl_toolkits.mplot3d import axes3d

        fig = pylab.figure()
        ax = axes3d.Axes3D(fig)
    
        # Construct matrices for r and c values
        rn,cn = mat.shape
        c = outer(ones(rn),arange(cn*1.0))
        r = outer(arange(rn*1.0),ones(cn))
    
        if type=="wireframe":
            ax.plot_wireframe(r,c,mat)
        elif type=="surface":
            # Sometimes fails for no obvious reason
            ax.plot_surface(r,c,mat)
        elif type=="contour":
            # Works but not usually very useful
            ax.contour3D(r,c,mat)
        else:
            raise ValueError("Unknown plot type "+str(type))
            
        ax.set_xlabel('R')
        ax.set_ylabel('C')
        ax.set_zlabel('Value')
    
        self._generate_figure(p)



def matrixplot3d_gnuplot(mat,title=None,outputfilename="tmp.ps"):
    """
    Simple plotting for any matrix as a 3D surface with axes.

    Currently requires the gnuplot-py package to be installed, plus
    the external gnuplot program; likely to be removed once Matplotlib
    supports 3D plots better.

    Unlikely to work on non-UNIX systems.

    Should return when it completes, but for some reason the Topographica
    prompt is not available until this command finishes.
    """
    import Gnuplot
    from os import system
    
    psviewer="gv" # Should be a parameter, or handled better somehow
    g = Gnuplot.Gnuplot(debug=0) #debug=1: output commands to stderr
    r,c = mat.shape
    x = arange(r*1.0)
    y = arange(c*1.0)
    # The .tolist() command is necessary to avoid bug in gnuplot-py,
    # which will otherwise convert a 2D float array into integers (!)
    m = numpy.asarray(mat,dtype="float32").tolist()
    #g("set parametric")
    g("set data style lines")
    g("set hidden3d")
    g("set xlabel 'R'")
    g("set ylabel 'C'")
    g("set zlabel 'Value'")
    if title: g.title(title)
    
    if outputfilename:
        g("set terminal postscript eps color solid 'Times-Roman' 14")
        g("set output '"+outputfilename+"'")
        g.splot(Gnuplot.GridData(m,x,y, binary=1))
        #g.hardcopy(outputfilename, enhanced=1, color=1)
        system(psviewer+" "+outputfilename+" &")

    else:
        g.splot(Gnuplot.GridData(m,x,y, binary=1))
        raw_input('Please press return to continue...\n')



class histogramplot(PylabPlotCommand):
    """
    Compute and plot the histogram of the supplied data.
    
    See help(pylab.hist) for help on the histogram function itself.

    If given, colors is an iterable collection of matplotlib.colors
    (see help (matplotlib.colors) ) specifying the bar colors.

    Example use:
        histogramplot([1,1,1,2,2,3,4,5],title='hist',colors='rgb',bins=3,normed=1)
    """
    
    # JABALERT: All but the first two arguments should probably be Parameters    
    def __call__(self,data,colors=None,**params):
        p=ParamOverrides(self,params,allow_extra_keywords=True)        
    
        pylab.figure(figsize=(4,2))
        n,bins,bars = pylab.hist(data,**(p.extra_keywords()))
    
        # if len(bars)!=len(colors), any extra bars won't have their
        # colors changed, or any extra colors will be ignored.
        if colors: [bar.set_fc(color) for bar,color in zip(bars,colors)]

        self._generate_figure(p)



class gradientplot(matrixplot):
    """
    Compute and show the gradient plot of the supplied data.
    Translated from Octave code originally written by Yoonsuck Choe.

    If the data is specified to be cyclic, negative differences will
    be wrapped into the range specified (1.0 by default).
    """
    
    # JABALERT: All but the first two arguments should probably be Parameters    
    def __call__(self,data,cyclic=True,cyclic_range=1.0,**params):
        p=ParamOverrides(self,params)
    
        r,c = data.shape
        dx = numpy.diff(data,1,axis=1)[0:r-1,0:c-1]
        dy = numpy.diff(data,1,axis=0)[0:r-1,0:c-1]
    
        if cyclic: # Wrap into the specified range
            # Convert negative differences to an equivalent positive value
            dx = wrap(0,cyclic_range,dx)
            dy = wrap(0,cyclic_range,dy)
            #
            # Make it increase as gradient reaches the halfway point,
            # and decrease from there
            dx = 0.5*cyclic_range-abs(dx-0.5*cyclic_range)
            dy = 0.5*cyclic_range-abs(dy-0.5*cyclic_range)

        super(gradientplot,self).__call__(sqrt(dx*dx+dy*dy),**p)



class fftplot(matrixplot):
    """
    Compute and show the 2D Fast Fourier Transform (FFT) of the supplied data.

    Example:: fftplot(topo.sim["V1"].sheet_views["OrientationPreference"].view()[0],filename="out")
    """
    
    def __call__(self,data,**params):
        p=ParamOverrides(self,params)
        fft_plot=1-abs(fftshift(fft2(data-0.5, s=None, axes=(-2,-1))))
        super(fftplot,self).__call__(fft_plot,**p)

        
class autocorrelationplot(matrixplot):
    """
    Compute and show the 2D autocorrelation of the supplied data.
    Requires the external SciPy package.
    
    Example:: autocorrelationplot(topo.sim["V1"].sheet_views["OrientationPreference"].view()[0],filename="out")
    """

    plot_type = param.Callable(default=pylab.autumn)
    
    def __call__(self,data,**params):
        p=ParamOverrides(self,params)
        import scipy.signal
        mat=scipy.signal.correlate2d(data,data)
        super(autocorrelationplot,self).__call__(mat,**p)


class activityplot(matrixplot):
    """
    Plots the activity in a sheet with axis labels in Sheet (not matrix) coordinates.

    Same as matrixplot, but only for matrices associated with a Sheet.
    By default plots the Sheet's activity, but any other matrix of the
    same size may be supplied for plotting in these coordinates instead.
    """
    def __call__(self,sheet,mat=None,**params):
        p=ParamOverrides(self,params)
        if p.extent is None: p.extent=sheet.bounds.aarect().lbrt()
        if mat is None: mat = sheet.activity
        super(activityplot,self).__call__(mat,**p)
        


class topographic_grid(PylabPlotCommand):
    """
    By default, plot the XPreference and YPreference preferences for all
    Sheets for which they are defined, using MatPlotLib.

    If sheet_views other than XPreference and YPreference are desired,
    the names of these can be passed in as arguments.
    """
    
    xsheet_view_name = param.String(default='XPreference',doc="""
        Name of the SheetView holding the X position locations.""")
    
    ysheet_view_name = param.String(default='YPreference',doc="""
        Name of the SheetView holding the Y position locations.""")
    
    axis = param.Parameter(default=[-0.5,0.5,-0.5,0.5],doc="""
        Four-element list of the plot bounds, i.e. [xmin, xmax, ymin, ymax].""")
    
    skip = param.Integer(default=1,bounds=[1,None],softbounds=[1,10],doc="""
        Plot every skipth line in each direction.  
        E.g. skip=4 means to keep only every fourth horizontal line
        and every fourth vertical line, except that the first and last
        are always included. The default is to include all data points.""")
    
    def __call__(self,**params):
        p=ParamOverrides(self,params)

        for sheet in topo.sim.objects(Sheet).values():
            if ((p.xsheet_view_name in sheet.sheet_views) and
                (p.ysheet_view_name in sheet.sheet_views)):
    
                x = sheet.sheet_views[p.xsheet_view_name].view()[0]
                y = sheet.sheet_views[p.ysheet_view_name].view()[0]
                
                pylab.figure(figsize=(5,5))
    
                # This one-liner works in Octave, but in matplotlib it
                # results in lines that are all connected across rows and columns,
                # so here we plot each line separately:
                #   pylab.plot(x,y,"k-",transpose(x),transpose(y),"k-")
                # Here, the "k-" means plot in black using solid lines;
                # see matplotlib for more info.
                isint=pylab.isinteractive() # Temporarily make non-interactive for plotting
                pylab.ioff()
                for r,c in zip(y[::p.skip],x[::p.skip]):
                    pylab.plot(c,r,"k-")
                for r,c in zip(transpose(y)[::p.skip],transpose(x)[::p.skip]):
                    pylab.plot(c,r,"k-")

                # Force last line avoid leaving cells open
                if p.skip != 1:
                    pylab.plot(x[-1],y[-1],"k-")
                    pylab.plot(transpose(x)[-1],transpose(y)[-1],"k-")
                
                pylab.xlabel('x')
                pylab.ylabel('y')
                # Currently sets the input range arbitrarily; should presumably figure out
                # what the actual possible range is for this simulation (which would presumably
                # be the maximum size of any GeneratorSheet?).
                pylab.axis(p.axis)
                p.title='Topographic mapping to '+sheet.name+' at time '+topo.sim.timestr()
    
                if isint: pylab.ion()
                p.filename_suffix="_"+sheet.name
                self._generate_figure(p)
            

class overlaid_plots(PylabPlotCommand):
    """
    Use matplotlib to make a plot combining a bitmap and line-based overlays.
    """
    
    plot_template = param.List(default=[{'Hue':'OrientationPreference'}],doc="""
        Template for the underlying bitmap plot.""")

    overlay = param.List(default=[('contours','OcularPreference',0.5,'black'),
                                ('arrows','DirectionPreference','DirectionSelectivity','white')],doc="""
        List of overlaid plots, where each list item may be a 4-tuple
        specifying either a contour line or a field of arrows::

          ('contours',map-name,contour-value,line-color)

          ('arrows',arrow-location-map-name,arrow-size-map-name,arrow-color)

        Any number or combination of contours and arrows may be supplied.""")

    normalize = param.Boolean(default='Individually',doc="""
        Type of normalization, if any, to use. Options include 'None',
        'Individually', and 'AllTogether'. See
        topo.plotting.plotgroup.TemplatePlotGroup.normalize for more
        details.""")


    def __call__(self,**params):
        p=ParamOverrides(self,params)

        for template in p.plot_template:
        
            for sheet in topo.sim.objects(Sheet).values():
                name=template.keys().pop(0)
                plot=make_template_plot(template,sheet.sheet_views,sheet.xdensity,sheet.bounds,p.normalize,name=template[name])        
                if plot:
                    bitmap=plot.bitmap
                    pylab.figure(figsize=(5,5))
                    isint=pylab.isinteractive() # Temporarily make non-interactive for plotting
                    pylab.ioff()                                         # Turn interactive mode off 
                                        
                    pylab.imshow(bitmap.image,origin='lower',interpolation='nearest')                           
                    pylab.axis('off')
    
                    for (t,pref,sel,c) in p.overlay:
                        v = pylab.flipud(sheet.sheet_views[pref].view()[0])
                        
                        if (t=='contours'):
                            pylab.contour(v,[sel,sel],colors=c,linewidths=2)
                            
                        if (t=='arrows'):                                                       
                            s = pylab.flipud(sheet.sheet_views[sel].view()[0])
                            scale=int(pylab.ceil(log10(len(v))))
                            X=pylab.array([x for x in xrange(len(v)/scale)])                  
                            v_sc=pylab.zeros((len(v)/scale,len(v)/scale))
                            s_sc=pylab.zeros((len(v)/scale,len(v)/scale))
                            for i in X:
                                for j in X:
                                    v_sc[i][j]=v[scale*i][scale*j]
                                    s_sc[i][j]=s[scale*i][scale*j]
                            pylab.quiver(scale*X,scale*X,-cos(2*pi*v_sc)*s_sc,-sin(2*pi*v_sc)*s_sc,color=c,edgecolors=c,minshaft=3,linewidths=1)                                                
                                                        
                    p.title='%s overlaid with %s at time %s' %(plot.name,pref,topo.sim.timestr())
                    if isint: pylab.ion()
                    p.filename_suffix="_"+sheet.name                    
                    self._generate_figure(p)
                


class tuning_curve(PylabPlotCommand):
    """
    Plot a tuning curve for a feature, such as orientation, contrast, or size.

    The curve datapoints are collected from the curve_dict for
    the units at the specified coordinates in the specified sheet
    (where the units and sheet may be set by a GUI, using 
    topo.analysis.featureresponses.UnitCurveCommand.sheet and 
    topo.analysis.featureresponses.UnitCurveCommand.coords, 
    or by hand).
    """
      
    coords = param.List(default=[(0,0)],doc="""
        List of coordinates of units to measure.""")

    sheet = param.ObjectSelector(
        default=None,doc="""
        Name of the sheet to use in measurements.""")

    x_axis = param.String(default="",doc="""
        Feature to plot on the x axis of the tuning curve""")

    # Can we list some alternatives here, if there are any
    # useful ones?
    plot_type = param.Callable(default=pylab.plot,doc="""
        Matplotlib command to generate the plot.""")

    unit = param.String(default="",doc="""
        String to use in labels to specify the units in which curves are plotted.""")

    legend=param.Boolean(default=True, doc="""
        Whether or not to include a legend in the plot.""")

    __abstract = True


    def _format_x_tick_label(self,x):
        return "%3.1f" % x

    def _rotate(self, seq, n=1):
        n = n % len(seq) # n=hop interval
        return seq[n:] + seq[:n]

    def _curve_values(self, i_value, j_value, curve):
        """Return the x, y, and x ticks values for the specified curve from the curve_dict"""
        x_values=sorted(curve.keys())
        y_values=[curve[key].view()[0][i_value,j_value] for key in x_values]
        return x_values,y_values,x_values

    def _reduce_ticks(self,ticks):
        x = [];
        y=  [];
        num_ticks = 5;
        y.append(ticks[0])
        x.append(0)
        for i in xrange(0,num_ticks):
            y.append(y[-1]+numpy.pi/(num_ticks+1));
            x.append(x[-1]+numpy.pi/(num_ticks+1));
        y.append(y[-1]+numpy.pi/(num_ticks+1));
        x.append(3.14)   
        return (x,y) 
                            

    def __call__(self,**params):
        p=ParamOverrides(self,params)
        sheet = p.sheet
        for coordinate in p.coords:
            i_value,j_value=sheet.sheet2matrixidx(coordinate[0],coordinate[1])
        
            pylab.figure(figsize=(7,7))
            isint=pylab.isinteractive()
            pylab.ioff()
    
            pylab.ylabel('Response',fontsize='large')
            pylab.xlabel('%s (%s)' % (p.x_axis.capitalize(),p.unit),fontsize='large')
            pylab.title('Sheet %s, coordinate(x,y)=(%0.3f,%0.3f) at time %s' % 
                        (sheet.name,coordinate[0],coordinate[1],topo.sim.timestr()))
            p.title='%s: %s Tuning Curve' % (topo.sim.name,p.x_axis.capitalize())
        
            self.first_curve=True
            for curve_label in sorted(sheet.curve_dict[p.x_axis].keys()):
                x_values,y_values,ticks=self._curve_values(i_value,j_value,sheet.curve_dict[p.x_axis][curve_label])
                
                x_tick_values,ticks = self._reduce_ticks(ticks)
                labels = [self._format_x_tick_label(x) for x in ticks]
                pylab.xticks(x_tick_values, labels,fontsize='large')
                pylab.yticks(fontsize='large')
                p.plot_type(x_values, y_values, label=curve_label,lw=3.0)
                self.first_curve=False
                 
            if isint: pylab.ion()
            if p.legend: pylab.legend(loc=2)
            self._generate_figure(p)



class cyclic_tuning_curve(tuning_curve):
    """
    Same as tuning_curve, but rotates the curve so that minimum y
    values are at the minimum x value to make the plots easier to
    interpret.  Such rotation is valid only for periodic quantities
    like orientation or direction, and only if the correct period
    is set.
    
    At present, the y_values and labels are rotated by an amount
    determined by the minmum y_value for the first curve plotted
    (usually the lowest contrast curve).
    """

    cyclic_range = param.Number(default=pi,bounds=(0,None),softbounds=(0,10),doc="""
        Range of the cyclic quantity (e.g. pi for the orientation of
        a symmetric stimulus, or 2*pi for motion direction or the
        orientation of a non-symmetric stimulus).""")
    
    unit = param.String(default="degrees",doc="""
        String to use in labels to specify the units in which curves are plotted.""")


    # This implementation should work for quantities periodic with
    # some multiple of pi that we want to express in degrees, but it
    # will need to be reimplemented in a subclass to work with other
    # cyclic quantities.
    def _format_x_tick_label(self,x):
        return str(int(180*x/pi))


    def _curve_values(self, i_value, j_value, curve):
        """
        Return the x, y, and x ticks values for the specified curve from the curve_dict.

        With the current implementation, there may be cases (i.e.,
        when the lowest contrast curve gives a lot of zero y_values)
        in which the maximum is not in the center.  This may
        eventually be changed so that the preferred orientation is in
        the center.
        """
        if self.first_curve==True:
            x_values= sorted(curve.keys())
            y_values=[curve[key].view()[0][i_value,j_value] for key in x_values]
            
            min_arg=argmin(y_values)
            x_min=x_values[min_arg] 
            y_min=y_values[min_arg]
            y_values=self._rotate(y_values, n=min_arg)
            self.ticks=self._rotate(x_values, n=min_arg)
            self.ticks+=[x_min]
            x_max=min(x_values)+self.cyclic_range
            x_values.append(x_max)
            y_values.append(y_min)

            self.x_values=x_values
        else:
            y_values=[curve[key].view()[0][i_value,j_value] for key in self.ticks]

        return self.x_values,y_values,self.ticks

            

def plot_cfproj_mapping(dest,proj='Afferent',style='b-'):
    """
    Given a CF sheet receiving a CFProjection, plot
    the mapping of the dests CF centers on the src sheet.
    """
    if isinstance(dest,str):
        from topo import sim
        dest = sim[dest]
    plot_coord_mapping(dest.projections()[proj].coord_mapper,
                       dest,style=style)


# JABALERT: not sure whether this is currently used
def plot_coord_mapping(mapper,sheet,style='b-'):
    """
    Plot a coordinate mapping for a sheet.
    
    Given a CoordinateMapperFn (as for a CFProjection) and a sheet
    of the projection, plot a grid showing where the sheet's units
    are mapped.
    """

    from pylab import plot,hold,ishold
    
    xs = sheet.sheet_rows()
    ys = sheet.sheet_cols()

    hold_on = ishold()
    if not hold_on:
        plot()
    hold(True)

    for y in ys:
        pts = [mapper(x,y) for x in xs]
        plot([u for u,v in pts],
             [v for u,v in pts],
             style)

    for x in xs:
        pts = [mapper(x,y) for y in ys]
        plot([u for u,v in pts],
             [v for u,v in pts],
             style)

    hold(hold_on)
    

# JABALERT: Untested as of Mon Nov 10 12:59:54 GMT 2008
class plot_tracked_attributes(PylabPlotCommand):
    """
    Plots parameter values associated with an AttributeTrackingTF.
    Example call:
    VT=AttributeTrackingTF(function=HE, debug_params=['a', 'b',], units=[(0,0),(1,1)], step=1)
    plot_tracked_attributes(VT,0,10000,attrib_names=['a'],units=[(0,0)], filename='V1')
    """

    # JABALERT: These parameters need to be documented.
    raw = param.Boolean(default=False)

    attrib_names = param.List(default=[])
    
    ylabel = param.String(default="")

    # Should be renamed to coords to match other commands
    units = param.List(default=[])

    ybounds = param.Parameter(default=(None,None))
   

    # JABALERT: All but the first two arguments should probably be Parameters
    def __call__(self,output_fn,init_time=0,final_time=None,**params):
        p=ParamOverrides(self,params)

        if final_time is None:
            final_time=topo.sim.time()
            
        attrs = p.attrib_names if len(p.attrib_names)>0 else output_fn.attrib_names
        for a in attrs:
            pylab.figure(figsize=(6,4))
            isint=pylab.isinteractive()
            pylab.ioff()
            pylab.grid(True)
            ylabel=p.ylabel
            pylab.ylabel(a+" "+ylabel)
            pylab.xlabel('Iteration Number')
            
            coords = p.units if len(p.units)>0 else output_fn.units
            for coord in coords:
                y_data=[y for (x,y) in output_fn.values[a][coord]]
                x_data=[x for (x,y) in output_fn.values[a][coord]]
                if p.raw==True:
                    plot_data=zip(x_data,y_data)
                    pylab.save(normalize_path(p.filename+a+'(%.2f, %.2f)' %(coord[0], coord[1])),plot_data,fmt='%.6f', delimiter=',')
                
                
                pylab.plot(x_data,y_data, label='Unit (%.2f, %.2f)' %(coord[0], coord[1]))
                (ymin,ymax)=p.ybounds
                pylab.axis(xmin=init_time,xmax=final_time,ymin=ymin,ymax=ymax)
                    
            if isint: pylab.ion()
            pylab.legend(loc=0)
            p.title=topo.sim.name+': '+a
            p.filename_suffix=a
            self._generate_figure(p)



# JABALERT: Should be updated to plot for a specified list of sheets,
# and then the combination of all of them, so that it will work for
# any network.  Will need to remove the simple_sheet_name and
# complex_sheet_name parameters once that works.
class plot_modulation_ratio(PylabPlotCommand):
    """
    This function computes the modulation ratios of neurons in the
    specified sheets and plots their histograms. See
    analysis.vision.complexity for more info.
    """

    # JABALERT: All but the first argument should probably be Parameters
    def __call__(self,fullmatrix,simple_sheet_name=None,complex_sheet_name=None,bins=frange(0,2.0,0.1,inclusive=True),**params):
        p=ParamOverrides(self,params)

        from topo.analysis.vision import complexity
        if (topo.sim.objects().has_key(simple_sheet_name) and topo.sim.objects().has_key(complex_sheet_name)):
            v1s = complexity(fullmatrix[topo.sim[simple_sheet_name]]).flatten()
            v1c = complexity(fullmatrix[topo.sim[complex_sheet_name]]).flatten()
            #double the number of complex cells to reflect large width of layer 2/3
            v1c = numpy.concatenate((array(v1c),array(v1c)),axis=1)
            pylab.figure()
            n = pylab.subplot(311)
            pylab.hist(v1s,bins)
            pylab.axis([0,2.0,0,4100])
	    n.yaxis.set_major_locator(matplotlib.ticker.MaxNLocator(3))
            
	    n = pylab.subplot(312)
            pylab.hist(v1c,bins)
            pylab.axis([0,2.0,0,4100])
	    n.yaxis.set_major_locator(matplotlib.ticker.MaxNLocator(3))
            
	    n = pylab.subplot(313)
            pylab.hist(numpy.concatenate((array(v1s),array(v1c)),axis=1),bins)
            pylab.axis([0,2.0,0,4100])
	    n.yaxis.set_major_locator(matplotlib.ticker.MaxNLocator(3))

        self._generate_figure(p)



class measure_position_pref(PositionMeasurementCommand):
    """Measure a position preference map by collating the response to patterns."""
    
    scale = param.Number(default=0.3)

    def _feature_list(self,p):
        width =1.0*p.x_range[1]-p.x_range[0]
        height=1.0*p.y_range[1]-p.y_range[0]
        return [Feature(name="x",range=p.x_range,step=width/p.divisions,preference_fn=self.preference_fn),
                Feature(name="y",range=p.y_range,step=height/p.divisions,preference_fn=self.preference_fn)]


from topo.misc.distribution import DSF_WeightedAverage
pg= create_plotgroup(name='Position Preference',category="Preference Maps",
           doc='Measure preference for the X and Y position of a Gaussian.',
           pre_plot_hooks=[measure_position_pref.instance(
            preference_fn=DSF_WeightedAverage( selectivity_scale=(0.,17.) ))],
           plot_hooks=[topographic_grid.instance()],
           normalize='Individually')

pg.add_plot('X Preference',[('Strength','XPreference')])
pg.add_plot('Y Preference',[('Strength','YPreference')])
pg.add_plot('Position Preference',[('Red','XPreference'),
                                   ('Green','YPreference')])



class measure_cog(ParameterizedFunction):
    """
    Calculate center of gravity (CoG) for each CF of each unit in each CFSheet.

    Unlike measure_position_pref and other measure commands, this one
    does not work by collating the responses to a set of input patterns.
    Instead, the CoG is calculated directly from each set of incoming
    weights.  The CoG value thus is an indirect estimate of what
    patterns the neuron will prefer, but is not limited by the finite
    number of test patterns as the other measure commands are.

    Measures only one projection for each sheet, as specified by the
    proj_name parameter.  The default proj_name of '' selects the
    first non-self connection, which is usually useful to examine for
    simple feedforward networks, but will not necessarily be useful in
    other cases.
    """
      
    proj_name = param.String(default='',doc="""
        Name of the projection to measure; the empty string means 'the first
        non-self connection available'.""")

    def __call__(self,**params):
        p=ParamOverrides(self,params)
        
        measured_sheets = [s for s in topo.sim.objects(CFSheet).values()
                           if hasattr(s,'measure_maps') and s.measure_maps]

        # Could easily be extended to measure CoG of all projections
        # and e.g. register them using different names (e.g. "Afferent
        # XCoG"), but then it's not clear how the PlotGroup would be
        # able to find them automatically (as it currently supports
        # only a fixed-named plot).
        requested_proj=p.proj_name
        for sheet in measured_sheets:
            for proj in sheet.in_connections:
                if (proj.name == requested_proj) or \
                   (requested_proj == '' and (proj.src != sheet)):
                    self._update_proj_cog(proj)
                    if requested_proj=='':
                        print "measure_cog: Measured %s projection %s from %s" % \
                              (proj.dest.name,proj.name,proj.src.name)
                        break


    def _update_proj_cog(self,proj):
        """Measure the CoG of the specified projection and register corresponding SheetViews."""
        
        sheet=proj.dest
        rows,cols=sheet.activity.shape
        xpref=zeros((rows,cols),Float)
        ypref=zeros((rows,cols),Float)
        
        for r in xrange(rows):
            for c in xrange(cols):
                cf=proj.cfs[r,c]
                r1,r2,c1,c2 = cf.input_sheet_slice
                row_centroid,col_centroid = centroid(cf.weights)
                xcentroid, ycentroid = proj.src.matrix2sheet(
                        r1+row_centroid+0.5,
                        c1+col_centroid+0.5)
            
                xpref[r][c]= xcentroid
                ypref[r][c]= ycentroid
            
                sheet.sheet_views['XCoG']=SheetView((xpref,sheet.bounds), sheet.name,
                                                    sheet.precedence,topo.sim.time(),sheet.row_precedence)
                
                sheet.sheet_views['YCoG']=SheetView((ypref,sheet.bounds), sheet.name,
                                                    sheet.precedence,topo.sim.time(),sheet.row_precedence)


pg= create_plotgroup(name='Center of Gravity',category="Preference Maps",
             doc='Measure the center of gravity of each ConnectionField in a Projection.',
             pre_plot_hooks=[measure_cog.instance()],
             plot_hooks=[topographic_grid.instance(xsheet_view_name="XCoG",ysheet_view_name="YCoG")],
             normalize='Individually')
pg.add_plot('X CoG',[('Strength','XCoG')])
pg.add_plot('Y CoG',[('Strength','YCoG')])
pg.add_plot('CoG',[('Red','XCoG'),('Green','YCoG')])


class measure_or_tuning_fullfield(FeatureCurveCommand):
    """
    Measures orientation tuning curve(s) of a particular unit using a
    full-field sine grating stimulus.  

    The curve can be plotted at various different values of the
    contrast (or actually any other parameter) of the stimulus.  If
    using contrast and the network contains an LGN layer, then one
    would usually specify michelson_contrast as the
    contrast_parameter. If there is no explicit LGN, then scale
    (offset=0.0) can be used to define the contrast.  Other relevant
    contrast definitions (or other parameters) can also be used,
    provided they are defined in PatternPresenter and the units
    parameter is changed as appropriate.
    """

    coords = param.Parameter(default=None,doc="""Ignored; here just to suppress warning.""")
    
    pattern_presenter = param.Callable(
        default=PatternPresenter(pattern_generator=SineGrating(),
                                 contrast_parameter="michelson_contrast"))

    
create_plotgroup(template_plot_type="curve",name='Orientation Tuning Fullfield',category="Tuning Curves",doc="""
            Plot orientation tuning curves for a specific unit, measured using full-field sine gratings.
            Although the data takes a long time to collect, once it is ready the plots
            are available immediately for any unit.""",
        pre_plot_hooks=[measure_or_tuning_fullfield.instance()],
        plot_hooks=[cyclic_tuning_curve.instance(x_axis="orientation")])



class measure_or_tuning(UnitCurveCommand):
    """
    Measures orientation tuning curve(s) of a particular unit.

    Uses a circular sine grating patch as the stimulus on the
    retina. 

    The curve can be plotted at various different values of the
    contrast (or actually any other parameter) of the stimulus.  If
    using contrast and the network contains an LGN layer, then one
    would usually specify weber_contrast as the contrast_parameter. If
    there is no explicit LGN, then scale (offset=0.0) can be used to
    define the contrast.  Other relevant contrast definitions (or
    other parameters) can also be used, provided they are defined in
    PatternPresenter and the units parameter is changed as
    appropriate.
    """

    num_orientation = param.Integer(default=12)

    static_parameters = param.List(default=["size","x","y"])

    def __call__(self,**params):
        p=ParamOverrides(self,params)
        self.params('sheet').compute_default()
        sheet=p.sheet
        
        for coord in p.coords:
            self.x=self._sheetview_unit(sheet,coord,'XPreference',default=coord[0])
            self.y=self._sheetview_unit(sheet,coord,'YPreference',default=coord[1])
            self._compute_curves(p,sheet)
               

create_plotgroup(template_plot_type="curve",name='Orientation Tuning',category="Tuning Curves",doc="""
            Measure orientation tuning for a specific unit at different contrasts,
            using a pattern chosen to match the preferences of that unit.""",
        pre_plot_hooks=[measure_or_tuning.instance()],
        plot_hooks=[cyclic_tuning_curve.instance(x_axis="orientation")],
        prerequisites=['XPreference'])



# JABALERT: Is there some reason not to call it measure_size_tuning?
class measure_size_response(UnitCurveCommand):
    """
    Measure receptive field size of one unit of a sheet.

    Uses an expanding circular sine grating stimulus at the preferred
    orientation and retinal position of the specified unit.
    Orientation and position preference must be calulated before
    measuring size response.

    The curve can be plotted at various different values of the
    contrast (or actually any other parameter) of the stimulus.  If
    using contrast and the network contains an LGN layer, then one
    would usually specify weber_contrast as the contrast_parameter. If
    there is no explicit LGN, then scale (offset=0.0) can be used to
    define the contrast.  Other relevant contrast definitions (or
    other parameters) can also be used, provided they are defined in
    PatternPresenter and the units parameter is changed as
    appropriate.
    """
    size=None # Disabled unused parameter

    static_parameters = param.List(default=["orientation","x","y"])

    num_sizes = param.Integer(default=10,bounds=(1,None),softbounds=(1,50),
                              doc="Number of different sizes to test.")

    max_size = param.Number(default=1.0,bounds=(0.1,None),softbounds=(1,50),
                              doc="Maximum extent of the grating")

    x_axis = param.String(default='size',constant=True)


    def __call__(self,**params):
        p=ParamOverrides(self,params)
        self.params('sheet').compute_default()
        sheet=p.sheet

        for coord in p.coords:
            # Orientations are stored as a normalized value beween 0
            # and 1, so we scale them by pi to get the true orientations.
            self.orientation=pi*self._sheetview_unit(sheet,coord,'OrientationPreference')            
            self.x=self._sheetview_unit(sheet,coord,'XPreference',default=coord[0])
            self.y=self._sheetview_unit(sheet,coord,'YPreference',default=coord[1])
            self._compute_curves(p,sheet)


    # Why not vary frequency too?  Usually it's just one number, but it could be otherwise.
    def _feature_list(self,p):
        return [Feature(name="phase",range=(0.0,2*pi),step=2*pi/p.num_phase,cyclic=True),
                Feature(name="frequency",values=p.frequencies),
                Feature(name="size",range=(0.0,self.max_size),step=self.max_size/p.num_sizes,cyclic=False)]


create_plotgroup(template_plot_type="curve",name='Size Tuning',category="Tuning Curves",
        doc='Measure the size preference for a specific unit.',
        pre_plot_hooks=[measure_size_response.instance()],
        plot_hooks=[tuning_curve.instance(x_axis="size",unit="Diameter of stimulus")],
        prerequisites=['OrientationPreference','XPreference'])



class measure_contrast_response(UnitCurveCommand):
    """
    Measures contrast response curves for a particular unit.

    Uses a circular sine grating stimulus at the preferred
    orientation and retinal position of the specified unit.
    Orientation and position preference must be calulated before
    measuring contrast response.

    The curve can be plotted at various different values of the
    contrast (or actually any other parameter) of the stimulus.  If
    using contrast and the network contains an LGN layer, then one
    would usually specify weber_contrast as the contrast_parameter. If
    there is no explicit LGN, then scale (offset=0.0) can be used to
    define the contrast.  Other relevant contrast definitions (or
    other parameters) can also be used, provided they are defined in
    PatternPresenter and the units parameter is changed as
    appropriate.
    """

    static_parameters = param.List(default=["size","x","y"])

    contrasts = param.List(class_=int,default=[10,20,30,40,50,60,70,80,90,100])

    relative_orientations = param.List(class_=float,default=[0.0, pi/6, pi/4, pi/2])
    
    x_axis = param.String(default='contrast',constant=True)

    units = param.String(default=" rad")

    def __call__(self,**params):
        p=ParamOverrides(self,params)
        self.params('sheet').compute_default()
        sheet=p.sheet

        for coord in p.coords:
            orientation=pi*self._sheetview_unit(sheet,coord,'OrientationPreference')
            self.curve_parameters=[{"orientation":orientation+ro} for ro in self.relative_orientations]

            self.x=self._sheetview_unit(sheet,coord,'XPreference',default=coord[0])
            self.y=self._sheetview_unit(sheet,coord,'YPreference',default=coord[1])
            
            self._compute_curves(p,sheet,val_format="%.4f")

    def _feature_list(self,p):
        return [Feature(name="phase",range=(0.0,2*pi),step=2*pi/p.num_phase,cyclic=True),
                Feature(name="frequency",values=p.frequencies),
                Feature(name="contrast",values=p.contrasts,cyclic=False)]


create_plotgroup(template_plot_type="curve",name='Contrast Response',category="Tuning Curves",
        doc='Measure the contrast response function for a specific unit.',
        pre_plot_hooks=[measure_contrast_response.instance()],
        plot_hooks=[tuning_curve.instance(x_axis="contrast",unit="%")],
        prerequisites=['OrientationPreference','XPreference'])



class measure_orientation_contrast(UnitCurveCommand):
    """
    Measures the response to a center sine grating disk and a surround
    sine grating ring at different contrasts of the central disk. 

    The central disk is set to the preferred orientation of the unit
    to be measured. The surround disk orientation (relative to the
    central grating) and contrast can be varied, as can the size of
    both disks.
    """

    pattern_presenter = param.Callable(
        default=PatternPresenter(pattern_generator=OrientationContrast(),
                                 contrast_parameter="weber_contrast"))

    size=None # Disabled unused parameter
    # Maybe instead of the below, use size and some relative parameter, to allow easy scaling?

    sizecenter=param.Number(default=0.5,bounds=(0,None),doc="""
        The size of the central pattern to present.""")

    sizesurround=param.Number(default=1.0,bounds=(0,None),doc="""
        The size of the surround pattern to present.""")

    thickness=param.Number(default=0.5,bounds=(0,None),softbounds=(0,1.5),doc="""Ring thickness.""")
    
    contrastsurround=param.Number(default=100,bounds=(0,100),doc="""Contrast of the surround.""")
    
    contrastcenter=param.Number(default=100,bounds=(0,100),doc="""Contrast of the center.""")
    
    x_axis = param.String(default='orientationsurround',constant=True)
    
    orientation_center = param.Number(default=0.0,softbounds=(0.0,numpy.pi),doc="""
        Orientation of the center grating patch""")

    units = param.String(default="%")

    static_parameters = param.List(default=["x","y","sizecenter","sizesurround","orientationcenter","thickness","contrastcenter"])

    curve_parameters=param.Parameter([{"contrastsurround":30},{"contrastsurround":60},{"contrastsurround":80},{"contrastsurround":90}],doc="""
        List of parameter values for which to measure a curve.""")

    or_surrounds = []
    
    def __call__(self,**params):
        p=ParamOverrides(self,params)
        self.params('sheet').compute_default()
        sheet=p.sheet
        for coord in p.coords:
            self.or_surrounds=[]
            orientation=p.orientation_center
            self.orientationcenter=orientation

            for i in xrange(0,self.num_orientation):
                self.or_surrounds.append(orientation+i*pi/(self.num_orientation))
            
            self.x=self._sheetview_unit(sheet,coord,'XPreference',default=coord[0])
            self.y=self._sheetview_unit(sheet,coord,'YPreference',default=coord[1])
            
            self._compute_curves(p,sheet)

    def _feature_list(self,p):
        return [Feature(name="phase",range=(0.0,2*pi),step=2*pi/p.num_phase,cyclic=True),
    	 	Feature(name="frequency",values=p.frequencies),
		Feature(name="orientationsurround",values=self.or_surrounds,cyclic=True)]

create_plotgroup(template_plot_type="curve",name='Orientation Contrast',category="Tuning Curves",
                 doc='Measure the response of one unit to a center and surround sine grating disk.',
                 pre_plot_hooks=[measure_orientation_contrast.instance()],
                 plot_hooks=[tuning_curve.instance(x_axis="orientationsurround",unit="%")],
                 prerequisites=['OrientationPreference','XPreference'])        



class test_measure(UnitCurveCommand):

    static_parameters = param.List(default=["size","x","y"])

    x_axis = param.String(default='contrast',constant=True)

    units = param.String(default=" rad")

    def __call__(self,**params):
        p=ParamOverrides(self,params)
        self.params('sheet').compute_default()
        sheet=p.sheet
        self.x = 0.0
	self.y = 0.0
        for coord in p.coords:
            self._compute_curves(p,sheet,val_format="%.4f")

    def _feature_list(self,p):
        return [Feature(name="orientation",values=[1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0],cyclic=True),
    		Feature(name="contrast",values=[100],cyclic=False)]

import types
__all__ = list(set([k for k,v in locals().items()
                    if isinstance(v,types.FunctionType) or 
                    (isinstance(v,type) and issubclass(v,ParameterizedFunction))
                    and not v.__name__.startswith('_')]))



"""
Sheet classes.

A Sheet is a two-dimensional arrangement of processing units,
typically modeling a neural region or a subset of cells in a neural
region.  Any new Sheet classes added to this directory will
automatically become available for any model.

$Id$
"""

# Imported here so that all Sheets will be in the same package
from topo.base.sheet import Sheet
from topo.base.projection import ProjectionSheet  # pyflakes:ignore (API import)
from topo.base.cf import CFSheet
from topo.misc.generatorsheet import GeneratorSheet

# Imported here for ease of access by users
from topo.base.boundingregion import BoundingBox  # pyflakes:ignore (API import)
from topo.base.sheet import activity_type  # pyflakes:ignore (API import)

import numpy

import topo

import param

from topo.base.cf import MaskedCFIter
from topo.base.projection import Projection
from topo.base.simulation import FunctionEvent, PeriodicEventSequence


class ActivityCopy(Sheet):
    """
    Copies incoming Activity patterns to its activity matrix and output port.

    Trivial Sheet class that is useful primarily as a placeholder for
    data that is computed elsewhere but that you want to appear as a
    Sheet, e.g. when wrapping an external simulation.
    """

    dest_ports=['Activity']
    src_ports=['Activity']
    
    def input_event(self,conn,data):
        self.input_data=data

    def process_current_time(self):
        if hasattr(self, 'input_data'):
            self.activity*=0
            self.activity+=self.input_data
            self.send_output(src_port='Activity',data=self.activity)
            del self.input_data



class SequenceGeneratorSheet(GeneratorSheet):
    """
    Sheet that generates a timed sequence of patterns.

    This sheet will repeatedly generate the input_sequence, with the
    given onsets.  The sequence is repeated every self.period time
    units.  If the total length of the sequence is longer than
    self.period, a warning is issued and the sequence repeats
    immediately after completion.
    """
    
    input_sequence = param.List(default=[],
          doc="""The sequence of patterns to generate.  Must be a list of
          (onset,generator) tuples. An empty list defaults to the
          single tuple: (0,self.input_generator), resulting in
          identical behavior to an ordinary GeneratorSheet.""")


    def __init__(self,**params):
        super(SequenceGeneratorSheet,self).__init__(**params)
        if not self.input_sequence:
            self.input_sequence = [(0,self.input_generator)]
            
    def start(self):
        assert self.simulation

        event_seq = []
        for delay,gen in self.input_sequence:
            event_seq.append(FunctionEvent(self.simulation.convert_to_time_type(delay),self.set_input_generator,gen))
            event_seq.append(FunctionEvent(0,self.generate))
        now = self.simulation.time()
        self.event = PeriodicEventSequence(now+self.simulation.convert_to_time_type(self.phase),self.simulation.convert_to_time_type(self.period),event_seq)
        self.simulation.enqueue_event(self.event)


def compute_joint_norm_totals(projlist,active_units_mask=True):
    """
    Compute norm_total for each CF in each projection from a group to
    be normalized jointly.
    """
    # Assumes that all Projections in the list have the same r,c size
    assert len(projlist)>=1
    iterator = MaskedCFIter(projlist[0],active_units_mask=active_units_mask)

    for junk,i in iterator():
        sums = [p.flatcfs[i].norm_total for p in projlist]
        joint_sum = numpy.add.reduce(sums)
        for p in projlist:
            p.flatcfs[i].norm_total=joint_sum


class JointNormalizingCFSheet(CFSheet):
    """
    A type of CFSheet extended to support joint sum-based normalization.

    For L1 normalization, joint normalization means normalizing the
    sum of (the absolute values of) all weights in a set of
    corresponding CFs in different Projections, rather than only
    considering weights in the same CF.

    This class provides a mechanism for grouping Projections (see
    _port_match and _grouped_in_projections) and a learn() function
    that computes the joint sums.  Joint normalization also requires
    having ConnectionField store and return a norm_total for each
    neuron, and having an TransferFn that will respect this norm_total
    rather than the strict total of the ConnectionField's weights.  At
    present, CFPOF_DivisiveNormalizeL1 and
    CFPOF_DivisiveNormalizeL1_opt do use norm_total; others can be
    extended to do something similar if necessary.

    To enable joint normalization, you can declare that all the
    incoming connections that should be normalized together each
    have a dest_port of:

    dest_port=('Activity','JointNormalize', 'AfferentGroup1'),

    Then all those that have this dest_port will be normalized
    together, as long as an appropriate TransferFn is being used.
    """

    joint_norm_fn = param.Callable(default=compute_joint_norm_totals,doc="""
        Function to use to compute the norm_total for each CF in each
        projection from a group to be normalized jointly.""")

    # JABALERT: Should check that whenever a connection is added to a
    # group, it has the same no of cfs as the existing connections.
    def start(self):
        self._normalize_weights(active_units_mask=False)        


    # CEBALERT: rename active_units_mask and default to False
    def _normalize_weights(self,active_units_mask=True):
        """
        Apply the weights_output_fns for every group of Projections.
        
        If active_units_mask is True, only active units will have
        their weights normalized.
        """    
        for key,projlist in self._grouped_in_projections('JointNormalize'):
            if key == None:
                normtype='Individually'
            else:
                normtype='Jointly'
                self.joint_norm_fn(projlist,active_units_mask)

            self.debug(normtype + " normalizing:")

            for p in projlist:
                p.apply_learn_output_fns(active_units_mask=active_units_mask)
                self.debug('  ',p.name)


    def learn(self):
        """
        Call the learn() method on every Projection to the Sheet, and
        call the output functions (jointly if necessary).
        """
        # Ask all projections to learn independently
        for proj in self.in_connections:
            if not isinstance(proj,Projection):
                self.debug("Skipping non-Projection "+proj.name)
            else:
                proj.learn()

        # Apply output function in groups determined by dest_port
        self._normalize_weights()

        

class JointNormalizingCFSheet_Continuous(JointNormalizingCFSheet):
    """
    CFSheet that runs continuously, with no 'resting' periods between pattern presentations.
    
    Note that learning occurs only when the time is a whole number.
    """
    def process_current_time(self):
        if self.new_input:
           self.new_input = False
           if(float(topo.sim.time()) % 1.0 == 0.0):
               #self.activate()
               if (self.plastic):
                   self.learn()
           #else:
           self.activate()



_public = list(set([_k for _k,_v in locals().items() if isinstance(_v,type) and issubclass(_v,Sheet)]))
_public += [
    "compute_joint_norm_totals",
    "BoundingBox",
    "ProjectionSheet",
    "activity_type",
]

# Automatically discover all .py files in this directory.
import os,fnmatch
__all__ = _public + [f.split('.py')[0] for f in os.listdir(__path__[0]) if fnmatch.fnmatch(f,'[!._]*.py')]
del f,os,fnmatch

# By default, avoid loading modules that rely on external libraries
# that might not be present on this system.
__all__.remove('ptztracker')



"""
Functions to test that map measurements haven't changed.

Use generate() to save data, and test() to check that the data is
unchanged in a later version. Results from generate() are already
checked into svn, so intentional changes to map measurement mean new
data must be generated and committed. E.g.

# ...deliberate change to orientation map measurement...
$ ./topographica -c 'default_density=8' examples/lissom_oo_or.ty -c \
  'topo.sim.run(100)' -c \
  'from topo.tests.test_map_measurement import *' -c \
  'generate(["Orientation Preference"])'
# now commit the resulting data file to the svn repository


plotgroups_to_test is a list of PlotGroups for which these
test functions are expected to be useful.
"""

import pickle

from numpy.testing import assert_array_almost_equal

import numpy
import topo

from param import resolve_path, normalize_path
from topo.command.analysis import *
from topo.command.pylabplot import *
from topo.plotting.plotgroup import plotgroups

# CEBALERT: change to be the all-in-one model eventually, and
# uncomment all ocular/disparity/direction groups below.
sim_name = 'lissom_oo_or'

# CEB: tests should store params of commands
# so we don't have to update data if someone
# e.g. edits a default value.

plotgroups_to_test = [
    # Several plotgroups are commented out because I was only thinking
    # about map measurement that results in sheet_views stored for V1.
    # (All could be included if the functions below were to be
    # extended to handle sheets other than V1, etc.)
    #'Activity',
    #'Connection Fields',
    #'Projection',
    #'Projection Activity',
    #'RF Projection',
    #'RF Projection (noise)',

    'Position Preference',
    # 'Center of Gravity',
    # 'Orientation and Ocular Preference',
    # 'Orientation and Direction Preference',
    # 'Orientation, Ocular and Direction Preference',
    'Orientation Preference',
    #'Ocular Preference',
    'Spatial Frequency Preference',
    #'PhaseDisparity Preference',
    'Orientation Tuning Fullfield',
    'Orientation Tuning',
    'Size Tuning',
    'Contrast Response',
    #'Retinotopy',  commented out because measurement is unstable
    #'Direction Preference',
    'Corner OR Preference'
    ]


def _reset_views(sheet):
    if hasattr(sheet,'sheet_views'):
        sheet.sheet_views = {}
    if hasattr(sheet,'curve_dict'):
        sheet.curve_dict = {}
    

def generate(plotgroup_names):
    assert topo.sim.name==sim_name
    assert topo.sim['V1'].nominal_density==8
    assert topo.sim.time()==100

    for name in plotgroup_names:
        print "* Generating data for plotgroups['%s']"%name

        views = {}
        sheet = topo.sim['V1']

        _reset_views(sheet)

        plotgroups[name]._exec_pre_plot_hooks()

        sheets_views = views[sheet.name] = {}

        if hasattr(sheet,'sheet_views'):
            sheets_views['sheet_views'] = sheet.sheet_views
        if hasattr(sheet,'curve_dict'):
            sheets_views['curve_dict'] = sheet.curve_dict

        filename = normalize_path('tests/%s_t%s_%s.data'%(sim_name,topo.sim.timestr(),
                                                          name.replace(' ','_')))
        print "Saving results to %s" % (filename)
        f = open(filename,'wb')
        pickle.dump((topo.version,views),f)
        f.close()

def checkclose(label,topo_version,x,y):
    errors=[]
    if not numpy.allclose(x,y,rtol=1e-05,atol=1e-07):
        print "...%s array is no longer close to the %s version:\n%s\n---\n%s" % (label,topo_version,x,y)
        errors=[label]
    else:
        print '...%s array is unchanged since data was generated (%s)' % (label,topo_version)
    return errors
    

def test(plotgroup_names):
    import topo
    import param
    assert topo.sim.name==sim_name
    assert topo.sim['V1'].nominal_density==8
    assert topo.sim.time()==100
    
    failing_tests=[]
    for name in plotgroup_names:
        print "\n* Testing plotgroups['%s']:"%name

        sheet = topo.sim['V1']
        _reset_views(sheet)
        plotgroups[name]._exec_pre_plot_hooks()

        filename = resolve_path('tests/%s_t%s_%s.data'%(sim_name,topo.sim.timestr(),
                                                          name.replace(' ','_')))
        print "Reading previous results from %s" % (filename)
        f = open(filename,'r')

        try:
            topo_version,previous_views = pickle.load(f)
        ########################################
        except AttributeError:
            # CEBALERT: code here just to support old data file. Should
            # generate a new one so it's no longer necessary.
            
            from topo.misc.legacy import preprocess_state

            import topo.base.boundingregion

            def _boundingregion_not_parameterized(instance,state):
                for a in ['initialized', '_name_param_value', 'nopickle']:
                    if a in state:
                        del state[a]

            preprocess_state(topo.base.boundingregion.BoundingRegion,
                             _boundingregion_not_parameterized)
    
            f.seek(0)
            topo_version,previous_views = pickle.load(f)
        ########################################
            
        f.close()

        if 'sheet_views' in previous_views[sheet.name]:
            previous_sheet_views = previous_views[sheet.name]['sheet_views']
            for view_name in previous_sheet_views:
                failing_tests += checkclose(sheet.name + " " + view_name,topo_version,
                                            sheet.sheet_views[view_name].view()[0],
                                            previous_sheet_views[view_name].view()[0])
                    
        if 'curve_dict' in previous_views[sheet.name]:
            previous_curve_dicts = previous_views[sheet.name]['curve_dict']
            # CB: need to cleanup var names e.g. val
            for curve_name in previous_curve_dicts:
                for other_param in previous_curve_dicts[curve_name]:
                    for val in previous_curve_dicts[curve_name][other_param]:
                        failing_tests += checkclose("%s %s %s %s" %(sheet.name,curve_name,other_param,val),topo_version,
                                                    sheet.curve_dict[curve_name][other_param][val].view()[0],
                                                    previous_curve_dicts[curve_name][other_param][val].view()[0])
                                          
    if failing_tests != []: raise AssertionError, "Failed map tests: %s" % (failing_tests)





"""
Classes providing GUI windows for PlotGroups, allowing sets of related plots
to be displayed.

$Id$
"""


import copy

import ImageTk

import Tkinter
from Tkinter import  Frame, TOP, YES, BOTH, X, LEFT, \
     RIGHT, DISABLED, NORMAL, Canvas, Label, NSEW, \
     NO, NONE,TclError

import param
# CEBALERT: here and elsewhere, change tk to paramtk
import paramtk as tk

from topo.base.sheet import Sheet
from topo.base.cf import CFSheet

from topo.command.pylabplot import matrixplot

from topo.misc.generatorsheet import GeneratorSheet

import topo

def with_busy_cursor(fn):
    """
    Decorator to show busy cursor for duration of fn call.
    """
    def busy_fn(widget,*args,**kw):
        if 'cursor' in widget.configure():
            old_cursor=widget['cursor']
            widget['cursor']='watch'
            widget.update_idletasks()

        try:
            fn(widget,*args,**kw)
        finally:
            # ensure old cursor is replaced even if fn() raises an
            # error
            if 'cursor' in widget.configure():
                widget['cursor']=old_cursor
            
    return busy_fn


class PlotGroupPanel(tk.TkParameterized,Frame):

    __abstract = True


    dock = param.Boolean(default=False,doc="on console or not")

    # Default size for images used on buttons
    button_image_size=(20,20)

    Refresh = tk.Button(image_path="tkgui/icons/redo-small.png",
        size=button_image_size,
        doc="""
        Refresh the current plot (i.e. force the current plot to be regenerated
        by executing pre_plot_hooks and plot_hooks).""")

    Redraw = tk.Button(image_path="tkgui/icons/redo-small.png",
        size=button_image_size,
        doc="""Redraw the plot from existing data (i.e. execute plot_hooks only).""")
    
    Enlarge = tk.Button(image_path="tkgui/icons/viewmag+_2.2.png",
        size=button_image_size,
        doc="""Increase the displayed size of the current plots by about 20%.""")
                          
    Reduce = tk.Button(image_path="tkgui/icons/viewmag-_2.1.png",
        size=button_image_size,doc="""
        Reduce the displayed size of the current plots by about 20%.
        A minimum size that preserves at least one pixel per unit is
        enforced, to ensure that no data is lost when displaying.""")

    Fwd = tk.Button(image_path="tkgui/icons/forward-2.0.png",
        size=button_image_size,doc="""
        Move forward through the history of all the plots shown in this window.""")

    Back = tk.Button(image_path="tkgui/icons/back-2.0.png",
        size=button_image_size,doc="""
        Move backward through the history of all the plots shown in
        this window.  When showing a historical plot, some functions
        will be disabled, because the original data is no longer
        available.""")
    
    gui_desired_maximum_plot_height = param.Integer(default=150,bounds=(0,None),doc="""
        Value to provide for PlotGroup.desired_maximum_plot_height for
        PlotGroups opened by the GUI.  Determines the initial, default
        scaling for the PlotGroup.""")
    
    # CB: is there a better way than using a property?
    def get_plotgroup(self):
        return self._extraPO
    def set_plotgroup(self,new_pg):
        self.change_PO(new_pg)
        
    plotgroup = property(get_plotgroup,set_plotgroup)


    def __init__(self,master,plotgroup,**params):
        """
        If your parameter should be available in history, add its name
        to the params_in_history list, otherwise it will be disabled
        in historical views.
        """
        
        tk.TkParameterized.__init__(self,master,extraPO=plotgroup,
                                    msg_handler=master.status,
                                    **params)
        Frame.__init__(self,master.content)

        self.parent = master #CEBALERT
        self.setup_plotgroup()





        self.canvases = []
        self.plot_labels = []

        ### JCALERT! Figure out why we need that!
        self._num_labels = 0
        
        self.plotgroups_history=[]
        self.history_index = 0
        self.params_in_history = [] # parameters valid to adjust in history
                                  
        # Factor for reducing or enlarging the Plots (where 1.2 = 20% change)
        self.zoom_factor = 1.2
        
        # CEBALERT: rename these frames
        self.control_frame_1 = Frame(master.noscroll)
        self.control_frame_1.pack(side=TOP,expand=NO,fill=X)

        self.control_frame_2 = Frame(master.noscroll)
        self.control_frame_2.pack(side=TOP,expand=NO,fill=X)

        self.plot_frame = Tkinter.LabelFrame(self,text=self.plotgroup.name)
        self.plot_frame.pack(side=TOP,expand=YES,fill=BOTH)#,padx=5,pady=5)        

        # CB: why did I need a new frame after switching to 8.5?
        # I've forgotten what i changed.
        self.plot_container = Tkinter.Frame(self.plot_frame)
        self.plot_container.pack(anchor="center")


        # Label does have a wraplength option...but it's in screen
        # units. Surely tk has a function to convert between
        # text and screen units?
        no_plot_note_text = """      
Press Refresh on the pre-plot hooks to generate the plot, after modifying the hooks below if necessary. Note that Refreshing may take some time.

Many hooks accept 'display=True' so that the progress can be viewed in an open Activity window, e.g. for debugging.
"""

        self.no_plot_note=Label(self.plot_container,text=no_plot_note_text,
                                justify="center",wraplength=350)
        self.no_plot_note_enabled=False


        self.control_frame_3 = Frame(master.noscroll_bottom)
        self.control_frame_3.pack(side=TOP,expand=NO,fill=X)

        self.control_frame_4 = Frame(self)
        self.control_frame_4.pack(side=TOP,expand=NO,fill=NONE)

        self.updatecommand_frame = Frame(self.control_frame_3)
        self.updatecommand_frame.pack(side=TOP,expand=YES,fill=X)

        self.plotcommand_frame = Frame(self.control_frame_3)
        self.plotcommand_frame.pack(side=TOP,expand=YES,fill=X)


        # CEBALERT: replace 
        self.messageBar = self.parent.status 

        self.pack_param('pre_plot_hooks',parent=self.updatecommand_frame,
                        expand='yes',fill='x',side='left')

        self.pack_param('Refresh',parent=self.updatecommand_frame,
                        on_set=self.refresh,side='right')
        self.params_in_history.append('Refresh')

        self.pack_param('plot_hooks',parent=self.plotcommand_frame,
                        expand='yes',fill='x',side='left')
        # CEBALERT: should disable unless data exists.
        self.pack_param('Redraw',parent=self.plotcommand_frame,
                        on_set=self.redraw_plots,side='right')
        
            
        self.pack_param('Enlarge',parent=self.control_frame_1,
                        on_set=self.enlarge_plots,side=LEFT)
        self.params_in_history.append('Enlarge') # CEBNOTE: while it's a GUI op

        self.pack_param('Reduce',parent=self.control_frame_1,
                        on_set=self.reduce_plots,side=LEFT)
        self.params_in_history.append('Reduce')


        if topo.tkgui.TK_SUPPORTS_DOCK:
            self.pack_param("dock",parent=self.control_frame_1,
                            on_set=self.set_dock,side=LEFT)


        # Don't need to add these two to params_in_history because their
        # availability is controlled separately (determined by what's
        # in the history)
        self.pack_param('Back',parent=self.control_frame_2,
                        on_set=lambda x=-1: self.navigate_pg_history(x),
                        side=LEFT)

        self.pack_param('Fwd',parent=self.control_frame_2,
                        on_set=lambda x=+1: self.navigate_pg_history(x),
                        side=LEFT)




        #################### RIGHT-CLICK MENU STUFF ####################
        ### Right-click menu for canvases; subclasses can add cascades
        ### or insert commands on the existing cascades.
        self._canvas_menu = tk.Menu(self, tearoff=0) #self.context_menu 

        self._unit_menu = tk.Menu(self._canvas_menu, tearoff=0)
        self._canvas_menu.add_cascade(menu=self._unit_menu,state=DISABLED,
                                      indexname='unit_menu')

        self._canvas_menu.add_separator()
        
        # CEBALERT: scheme for enabling/disabling menu items ('disable
        # items hack') needs to be generalized. What we have now is
        # just a mechanism to disable/enable cfs/rfs plots as
        # necessary. Hack includes the attribute below as well as
        # other items marked 'disable items hack'.
        # (Note that tk 8.5 has better handling of state switching
        # (using flags for each state, I think), so presumably this
        # can be cleaned up easily.)
        self._unit_menu_updaters = {}
        
        self._sheet_menu = tk.Menu(self._canvas_menu, tearoff=0)
        self._canvas_menu.add_cascade(menu=self._sheet_menu,state=DISABLED,
                                      indexname='sheet_menu')
        self._canvas_menu.add_separator()


        self.update_plot_frame(plots=False)

        #################################################################

        # CB: don't forget to include ctrl-q
        # import __main__; __main__.__dict__['qqq']=self


    def set_dock(self):
        if self.dock:
            topo.guimain.some_area.consume(self.parent)
            self.refresh_title()
        else:
            topo.guimain.some_area.eject(self.parent)
            self.refresh_title()
            


    def setup_plotgroup(self):
        """
        Perform any necessary initialization of the plotgroup.

        Subclasses can use this to set Parameters on their PlotGroups.
        """
        self.plotgroup.desired_maximum_plot_height=self.gui_desired_maximum_plot_height
        

    def __process_canvas_event(self,event,func):
        """
        Return a dictionary containing the event itself, and, if the
        event occurs on a plot of a sheet, store the plot and the
        coordinates ((r,c) and (x,y) for the cell center) on the sheet.

        Then, call func.
        """
        # CB: I want this to be called for all the canvas events - see
        # ALERT by canvas button bindings. Surely can do better than
        # just passing func through.
        plot=event.widget.plot
        event_info = {'event':event} # store event in case more info needed elsewhere

        # Later functions assume that if event_info does not contain
        # 'plot', then the event did not occur on a plot of a sheet.
        if plot.plot_src_name is not '':
            plot_width,plot_height=plot.bitmap.width(),plot.bitmap.height()
            if 0<=event.x<plot_width and 0<=event.y<plot_height:
                left,bottom,right,top=plot.plot_bounding_box.lbrt()
                # float() to avoid integer division
                x = (right-left)*float(event.x)/plot_width + left
                y = top - (top-bottom)*float(event.y)/plot_height 
                r,c = topo.sim[plot.plot_src_name].sheet2matrixidx(x,y)
                event_info['plot'] = plot
                event_info['coords'] = [(r,c),(x,y)]
            
        func(event_info)
        

    def _canvas_right_click(self,event_info,show_menu=True):
        """
        Update labels on right-click menu and popup the menu, plus store the event info
        for access by any menu commands that require it.

        If show_menu is False, popup menu is not displayed (in case subclasses
        wish to add extra menu items first).
        """        
        if 'plot' in event_info:
            plot = event_info['plot']

            self._canvas_menu.entryconfig("sheet_menu",
                label="Combined plot: %s %s"%(plot.plot_src_name,plot.name),
                state=NORMAL)            
            (r,c),(x,y) = event_info['coords']
            self._canvas_menu.entryconfig("unit_menu",
                label="Single unit:(% 3d,% 3d) Coord:(% 2.2f,% 2.2f)"%(r,c,x,y),
                state=NORMAL)
            self._right_click_info = event_info

            # CB: part of disable items hack
            for v in self._unit_menu_updaters.values(): v(plot)

            if show_menu:
                self._canvas_menu.tk_popup(event_info['event'].x_root,
                                           event_info['event'].y_root)


            

    def _update_dynamic_info(self,event_info):
        """
        Update dynamic information.
        """
        if 'plot' in event_info:
            plot = event_info['plot']
            (r,c),(x,y) = event_info['coords']
            location_string="%s Unit:(% 3d,% 3d) Coord:(% 2.2f,% 2.2f)"%(plot.plot_src_name,r,c,x,y)
            # CB: isn't there a nicer way to allow more info to be added?
            self.messageBar.dynamicinfo(self._dynamic_info_string(event_info,location_string))
        else:
            self.messageBar.dynamicinfo('')

        

        
    def _dynamic_info_string(self,event_info,x):
        """
        Subclasses can override to add extra relevant information.
        """
        return x


    # rename (not specific to plot_frame)
    # document, and make display_* methods semi-private methods
    def update_plot_frame(self,plots=True,labels=True,geom=False):
        """

        set geom True for any action that user would expect to lose
        his/her manual window size (e.g. pressing enlarge button)
        """
        
        if plots:
            self.plotgroup.scale_images()
            self.display_plots()
        if labels: self.display_labels()
        self.refresh_title()

        if len(self.canvases)==0:
            # CEB: check that pack's ok here
            self.no_plot_note.grid(row=1,column=0,sticky='nsew')
            self.no_plot_note_enabled=True
            self.representations['Enlarge']['widget']['state']=DISABLED
            self.representations['Reduce' ]['widget']['state']=DISABLED
            
        elif self.no_plot_note_enabled:
            self.no_plot_note.grid_forget()
            self.no_plot_note_enabled=False
            self.representations['Enlarge']['widget']['state']=NORMAL
            self.representations['Reduce' ]['widget']['state']=NORMAL

        self.__update_widgets_for_history()
        # have a general update_widgets method instead (that calls
        # update_widgets_for_history; can it also include
        # enlarge/reduce alterations?)

        # CBALERT: problem when docked: this event isn't being caught,
        # ie it doesn't end up going to the right place... (i.e. no
        # scrollbars when docked).
        #self.event_generate("<<SizeRight>>")
        self.parent.sizeright()
        if geom:
            try:
                self.parent.geometry('')
            except TclError:
                pass
        

    @with_busy_cursor
    def refresh_plots(self):
        """
        Call plotgroup's make_plots with update=True (i.e. run
        pre_plot_hooks and plot_hooks), then display the result.
        """
        self.plotgroup.make_plots(update=True)
        self.update_plot_frame()        
        self.add_to_plotgroups_history()


    @with_busy_cursor
    def redraw_plots(self):
        """
        Call plotgroup's make_plots with update=False (i.e. run only
        plot_hooks, not pre_plot_hooks), then display the result.
        """
        self.plotgroup.make_plots(update=False)
        self.update_plot_frame(labels=False)


    def rescale_plots(self):
        """
        Rescale the existing plots, without calling either the
        plot_hooks or the pre_plot_hooks, then display the result.
        """
        self.plotgroup.scale_images()
        self.update_plot_frame(labels=False,geom=True)

    
    def refresh(self,update=True):
        """
        Main steps for generating plots in the Frame. 

        # if update is True, the SheetViews are re-generated
        """
        
        # if we've been looking in the history, now need to return to the "current time"
        # plotgroup (but copy it: don't update the old one, which is a record of the previous state)
        if self.history_index!=0:
            self._switch_plotgroup(copy.copy(self.plotgroups_history[-1]))
            self.history_index = 0

        if update:
            self.refresh_plots()            
        else:
            self.redraw_plots()


    ### JABALERT: Can we make it simpler to make plots be put onto multiple lines?
    # (because this is just the smallest change I cpuld think of to
    # support row precedence, without altering the existing code!)
    def _determine_layout_of_plots(self,plots):
        """Calculate self._rows and self._cols, together giving the grid position of each plot."""
        distinct_precedences = sorted(set([p.row_precedence for p in plots]))

        # 2*i because labels will occupy odd rows
        precedence2row = dict([ (precedence,2*i)
                                for precedence,i in zip(distinct_precedences,
                                                        range(len(distinct_precedences)))])
        # CB: a 2d array might have been clearer...
        self._rows = [precedence2row[p.row_precedence] for p in plots]
        self._cols = []
        
        row_counts = dict([(row,0) for row in self._rows])        
        for row in self._rows:
            self._cols.append(row_counts[row])
            row_counts[row]+=1


    # CEBALERT: this method needs cleaning, along with its versions in subclasses.
    def display_plots(self):
        """

        This function should be redefined in subclasses for interesting
        things such as 2D grids.
        """
        plots = self.plotgroup.plots
        self._determine_layout_of_plots(plots)
        
        self.zoomed_images = [ImageTk.PhotoImage(p.bitmap.image) for p in plots]

        new_sizes = [(str(zi.width()),
                      str(zi.height()))
                     for zi in self.zoomed_images]
        old_sizes = [(canvas['width'],canvas['height'])
                     for canvas in self.canvases]

        # If the number of canvases or their sizes has changed, then
        # create a new set of canvases.  If the new images will fit into the
        # old canvases, reuse them (prevents flicker)
        


        if len(self.zoomed_images) != len(self.canvases) or \
               new_sizes != old_sizes:
            # Need new canvases...
            old_canvases = self.canvases
            self.canvases = [Canvas(self.plot_container,
                                    width=image.width(),
                                    height=image.height(),
                                    borderwidth=1,highlightthickness=0,
                                    relief='groove') 
                             for image in self.zoomed_images]
            for i,image,canvas in zip(range(len(self.zoomed_images)),
                                      self.zoomed_images,self.canvases):
                canvas.create_image(1,1,anchor="nw",image=image)
                canvas.grid(row=self._rows[i],column=self._cols[i],padx=5)

                
            for c in old_canvases:
                c.grid_forget()


        else:
            # Don't need new canvases...
            for i,image,canvas in zip(range(len(self.zoomed_images)),
                                      self.zoomed_images,self.canvases):
                canvas.create_image(1,1,anchor="nw",image=image)
                canvas.grid(row=self._rows[i],column=self._cols[i],padx=5)

        self._add_canvas_bindings()


    def _add_canvas_bindings(self):
        ### plotting over; bind events to each canvas
        for plot,canvas in zip(self.plotgroup.plots,self.canvases):
            # Store the corresponding plot with each canvas so that the
            # plot information (e.g. scale_factor) will be available
            # for the right_click menu.
            canvas.plot=plot
            # CEBALERT: I want process_canvas_event to be called for
            # all of these bindings, with an additional method also
            # called to do something specific to the action. I'm sure
            # python has something that lets this be done in a clearer
            # way.
            canvas.bind('<<right-click>>',lambda event: \
                        self.__process_canvas_event(event,self._canvas_right_click))
            canvas.bind('<Motion>',lambda event: \
                        self.__process_canvas_event(event,self._update_dynamic_info))

            canvas.bind('<Leave>',lambda event: \
                        self.__process_canvas_event(event,self._update_dynamic_info))
            # When user has a menu up, it's often natural to click
            # elsewhere to make the menu disappear. Need to update the
            # dynamic information in that case. (Happens on OS X
            # anyway, but needed on Win and linux.)
            canvas.bind('<Button-1>',lambda event: \
                        self.__process_canvas_event(event,self._update_dynamic_info))

        

        
        

    def display_labels(self):
        """
        This function should be redefined by subclasses to match any
        changes made to display__plots().  Depending on the situation,
        it may be useful to make this function a stub, and display the
        labels at the same time the images are displayed.
        """
        
        if len(self.canvases) == 0:
            pass
        elif self._num_labels != len(self.canvases):
            old_labels = self.plot_labels
            self.plot_labels = [Label(self.plot_container,text=each)
                                 for each in self.plotgroup.labels]
            for i in range(len(self.plot_labels)):
                self.plot_labels[i].grid(row=self._rows[i]+1,column=self._cols[i],sticky=NSEW)
            for l in old_labels:
                l.grid_forget()
            self._num_labels = len(self.canvases)
        else:  # Same number of labels; reuse to avoid flickering.
            for i in range(len(self.plot_labels)):
                self.plot_labels[i].configure(text=self.plotgroup.labels[i]) 

      

    # CEBERRORALERT (minor): if no plot's displayed and I click
    # enlarge, then the enlarge button gets disabled. If I then press
    # refresh to get a plot, I can't enlarge it because the button's
    # disabled. Probably need to reset button status if the plots
    # change.
    def reduce_plots(self):
        """Function called by widget to reduce the plot size, when possible."""
        if (not self.plotgroup.scale_images(1.0/self.zoom_factor)):
            self.representations['Reduce']['widget']['state']=DISABLED
        self.representations['Enlarge']['widget']['state']=NORMAL
        self.update_plot_frame(labels=False,geom=True)

    def enlarge_plots(self):
        """Function called by widget to increase the plot size, when possible."""
        if (not self.plotgroup.scale_images(self.zoom_factor)):
            self.representations['Enlarge']['widget']['state']=DISABLED
        self.representations['Reduce']['widget']['state']=NORMAL
        self.update_plot_frame(labels=False,geom=True)
        

######################################################################
### HISTORY METHODS

    # CEBERRORALERT: history grows and grows! Consider what happens when
    # a window's open with auto-refresh and many plots are generated
    # (e.g. measure_rfs). And plotgroups might be much bigger than they
    # need to be.

    # CEBALERT: in a history research, a disabled widget does not display
    # up-to-date information (e.g. normalize checkbutton doesn't change).
    def add_to_plotgroups_history(self):
        """
        If there are plots on display, and we're not doing a history research,
        the plotgroup is stored in the history.
        """
        if self.history_index==0 and not len(self.canvases)==0:
            self.plotgroups_history.append(copy.copy(self.plotgroup))
        self.__update_widgets_for_history() 

    def __set_widget_state(self,widget,state):
        # sets the widget's state to state, unless state=='normal'
        # and the widget's current state is 'readonly', in which
        # case readonly is preserved.
        # If a widget state was set to 'disabled' deliberately, this
        # will have the unwanted effect of enabling that widget.
        # Surely there's a better way than this!
        # (Probably the history stuff should store the old state
        # on the widget somewhere. That would also eliminate the
        # combobox-specific hack.)

        # CEBALERT: I guess some widgets don't have state?
        try:
            current_state = widget.configure('state')[3]  # pyflakes:ignore (try/except test)
        except TclError:
            return

        ### hack to deal with combobox: see tkparameterizedobject's
        ###  create_selector_widget().
        if state=='normal':
            if hasattr(widget,'_readonly_'):
                state='readonly'
        ###########################################################
                            
        widget.configure(state=state)
            
        
    def __update_widgets_for_history(self):
        """
        The plotgroup's non-history widgets are all irrelevant when the plotgroup's from
        history.
        """
        if self.history_index!=0: 
            state= 'disabled'
        else:
            state = 'normal'
        
        widgets_to_update = [self.representations[p_name]['widget']
                             for p_name in self.representations
                             if p_name not in self.params_in_history]

        for widget in widgets_to_update:
            self.__set_widget_state(widget,state)
            
        self.__update_history_buttons()


    def __update_history_buttons(self):
        """
        Enable/disable the back and forward buttons depending on
        where we are in a history research.
        """
        space_back = len(self.plotgroups_history)+self.history_index-1
        space_fwd  = -self.history_index

        back_button = self.representations['Back']['widget']
        forward_button = self.representations['Fwd']['widget']
        
        if space_back>0:
            back_button['state']='normal'
        else:
            back_button['state']='disabled'

        if space_fwd>0:
            forward_button['state']='normal'
        else:
            forward_button['state']='disabled'

    # JLENHANCEMENT: It would be nice to be able to scroll back through many
    # iterations.  Could put in a box for entering either the iteration
    # number you want to view, or perhaps how many you want to jump...
    def navigate_pg_history(self,steps):
        self.history_index+=steps
        self._switch_plotgroup(self.plotgroups_history[len(self.plotgroups_history)+self.history_index-1])
        self.update_plot_frame()

######################################################################


    def _switch_plotgroup(self,newpg):
        """
        Switch to a different plotgroup, e.g. one from the history buffer.
        Preserves some attributes from the current plotgroup that can apply
        across history, but leaves the others as-is.
        """
        oldpg=self.plotgroup
        
        newpg.desired_maximum_plot_height=oldpg.desired_maximum_plot_height
        newpg.sheet_coords=oldpg.sheet_coords
        newpg.integer_scaling=oldpg.integer_scaling
        
        self.plotgroup=newpg


###########################################################         



    def _plot_title(self):
        """
        Provide a string describing the current set of plots.

        Override in subclasses to provide more information.
        """
        return "%s at time %s"%(self.plotgroup.name,topo.sim.timestr(self.plotgroup.time))

                
    # rename to refresh_titles
    def refresh_title(self):
        """
        Set Window title and plot frame's title.
        """
        title = self._plot_title()
        
        self.plot_frame.configure(text=title)
        self.parent.title(str(topo.sim.name)+": "+title)
        # JABALERT: Used to say .replace(" at time ","/"); was there a reason?
    
    def destroy(self):
        """overrides toplevel destroy, adding removal from autorefresh panels"""
        if self in topo.guimain.auto_refresh_panels:
            topo.guimain.auto_refresh_panels.remove(self)
        Frame.destroy(self)
            


class SheetPanel(PlotGroupPanel):

    sheet_type = Sheet

    @classmethod
    def valid_context(cls):
        """
        Return true if there appears to be data available for this type of plot.

        To avoid confusing error messages, this method should be
        defined to return False in the case where there is no
        appropriate data to plot.  This information can be used to,
        e.g., gray out the appropriate menu item.
        By default, PlotPanels are assumed to be valid only for
        simulations that contain at least one Sheet.  Subclasses with
        more specific requirements should override this method with
        something more appropriate.
        """
        if topo.sim.objects(cls.sheet_type).items():
            return True
        else:
            return False
        

    def __init__(self,master,plotgroup,**params):
        super(SheetPanel,self).__init__(master,plotgroup,**params)

        self.pack_param('auto_refresh',parent=self.control_frame_1,
                        on_set=self.set_auto_refresh,
                        side=RIGHT)
        self.params_in_history.append('auto_refresh')

        if self.auto_refresh:
            topo.guimain.auto_refresh_panels.append(self)


        self.pack_param('normalize',parent=self.control_frame_1,
                        on_set=self.redraw_plots,side="right")
        self.pack_param('integer_scaling',parent=self.control_frame_2,
                        on_set=self.rescale_plots,side='right')
        self.pack_param('sheet_coords',parent=self.control_frame_2,
                        on_set=self.rescale_plots,side='right')

        self.params_in_history.append('sheet_coords')
        self.params_in_history.append('integer_scaling')


        
        self._unit_menu.add_command(label='Connection Fields',indexname='connection_fields',
                                    command=self._connection_fields_window)
                                    
        self._unit_menu.add_command(label='Receptive Field',
                                    indexname='receptive_field',
                                    command=self._receptive_field_window)

        self._unit_menu.add_command(label='Orientation Tuning Curves',
                                    indexname='or tuning curve',
                                    command=self._or_tuning_curve_window)

###### part of disable items hack #####
        self._unit_menu_updaters['connection_fields'] = self.check_for_cfs
        self._unit_menu_updaters['receptive_field'] = self.check_for_rfs        

    def check_for_cfs(self,plot):
        show_cfs = False
        if plot.plot_src_name in topo.sim.objects():
            if isinstance(topo.sim[plot.plot_src_name],CFSheet):
                show_cfs = True
        self.__showhide("connection_fields",show_cfs)

    def check_for_rfs(self,plot):
        show_rfs = False
        if plot.plot_src_name in topo.sim.objects():
            sheet = topo.sim[plot.plot_src_name]

            # RFHACK: if any one generator has RF views for this sheet, then enable the menu option
            # At the moment, just a hack to prevent menu option for generator sheets.
            if not isinstance(sheet,GeneratorSheet):
                show_rfs = True
            else:
                show_rfs = False

        self.__showhide("receptive_field",show_rfs)

    def __showhide(self,name,show):
        if show:
            state = 'normal'            
        else:
            state = 'disabled'
        self._unit_menu.entryconfig(name,state=state)
#######################################


    def set_auto_refresh(self):
        """
        Add or remove this panel from the console's
        auto_refresh_panels list.
        """
        if self.auto_refresh: 
            if not (self in topo.guimain.auto_refresh_panels):
                topo.guimain.auto_refresh_panels.append(self)
        else:
            if self in topo.guimain.auto_refresh_panels:
                topo.guimain.auto_refresh_panels.remove(self)



    def _connection_fields_window(self):
        """
        Open a Connection Fields plot for the unit currently
        identified by a right click.
        """
        if 'plot' in self._right_click_info:
            sheet = topo.sim[self._right_click_info['plot'].plot_src_name]
            # CEBERRORALERT: should avoid requesting cf out of range.
            center_x,center_y = self._right_click_info['coords'][1]
            topo.guimain['Plots']["Connection Fields"](x=center_x,y=center_y,sheet=sheet)                     

            
    def _receptive_field_window(self):
        """
        Open a Receptive Field plot for the unit currently
        identified by a right click.
        """
        if 'plot' in self._right_click_info:
            plot = self._right_click_info['plot']
            sheet = topo.sim[plot.plot_src_name]
            center_x,center_y=self._right_click_info['coords'][1]
            r,c = self._right_click_info['coords'][0]

            # RFHACK:
            # just matrixplot for whatever generators have the views
            for g in topo.sim.objects(GeneratorSheet).values():
                try:
                    view=g.sheet_views[('RFs',sheet.name,center_x,center_y)]
                    matrixplot(view.view()[0],
                               title=("Receptive Field of %s unit (%d,%d) at coord (%3.0f, %3.0f) at time %s" %
                                      (sheet.name,r,c,center_x,center_y,topo.sim.timestr(view.timestamp))))

                except KeyError:
                    # maybe lose this warning
                    topo.sim.warning("No RF measurements are available yet for input_sheet %s; run the Receptive Field plot for that input_sheet to see the RF."%g.name)


    def _or_tuning_curve_window(self):
        """
        Open a Tuning Curve plot for the unit currently
        identified by a right click.
        """
        if 'plot' in self._right_click_info:
            plot = self._right_click_info['plot']
            sheet = topo.sim[plot.plot_src_name]
            center_x,center_y=self._right_click_info['coords'][1]
            r,c = self._right_click_info['coords'][0]

            try:
                from topo.command.pylabplot import cyclic_tuning_curve
                cyclic_tuning_curve(x_axis="orientation",coords=[(center_x,center_y)],sheet=sheet)
            except AttributeError:
                topo.sim.warning("No orientation tuning curve measurements are available yet for sheet %s; run the Orientation Tuning (Fullfield) command and try again."%sheet.name)


    def conditional_refresh(self):
        """
        Only calls refresh() if auto_refresh is enabled.
        """
        if self.auto_refresh:self.refresh()

    def conditional_redraw(self):
        """
        Only calls redraw_plots() if auto_refresh is enabled.
        """
        if self.auto_refresh:self.redraw_plots()



"""
Classes for using robotic or other hardware using Topographica.

This module contains several classes for constructing robotics
interfaces to Topographica simulations.  It includes modules that read
input from or send output to robot devices, and a (quasi) real-time
simulation object that attempts to maintain a correspondence between
simulation time and real time.

This module requires the PlayerStage robot interface system (from
playerstage.sourceforge.net), and the playerrobot module for
high-level communications with Player robots.

$Id$
"""


import Image,ImageOps
from math import pi,cos,sin

import param

from topo.base.simulation import EventProcessor
from topo.pattern.image import GenericImage

from playerrobot import CameraDevice,PTZDevice


class CameraImage(GenericImage):
    """
    An image pattern generator that gets its image from a Player
    camera device.
    """
    
    camera = param.ClassSelector(CameraDevice,default=None,doc="""
       An instance of playerrobot.CameraDevice to be used
       to generate images.""")

    def __init__(self,**params):
        super(CameraImage,self).__init__(**params)
        self._image = None

    def _get_image(self,params):
        self._decode_image(*self.camera.image)
        return True

    def _decode_image(self,fmt,w,h,bpp,fdiv,data):
        if fmt==1:
            self._image = Image.new('L',(w,h))
            self._image.fromstring(data,'raw')
        else:
            # JPALERT: if not grayscale, then assume color.  This
            # should be expanded for other modes.
            rgb_im = Image.new('RGB',(w,h))
            rgb_im.fromstring(data,'raw')
            self._image = ImageOps.grayscale(rgb_im)
        


class CameraImageQueued(CameraImage):
    """
    A version of CameraImage that gets the image from the camera's image queue,
    rather than directly from the camera object.  Using queues is
    necessary when running the playerrobot in a separate process
    without shared memory.  When getting an image, this pattern
    generator will fetch every image in the image queue and use the
    most recent as the current pattern.
    """

    def _get_image(self,params):

        im_spec = None
        
        if self._image is None:
            # if we don't have an image then block until we get one
            im_spec = self.camera.image_queue.get()
            self.camera.image_queue.task_done()

        # Make sure we clear the image queue and get the most recent image.
        while not self.camera.image_queue.empty():
            im_spec = self.camera.image_queue.get_nowait()
            self.camera.image_queue.task_done()

        if im_spec:
            # If we got a new image from the queue, then
            # construct a PIL image from it.
            self._decode_image(*im_spec)
            return True
        else:
            return False


class PTZ(EventProcessor):
    """
    Pan/Tilt/Zoom control.

    This event processor takes input events on its 'Saccade' input
    port in the form of (amplitude,direction) saccade commands (as
    produced by the topo.sheet.saccade.SaccadeController class) and
    appropriately servoes the attached PTZ object.  There is not
    currently any dynamic zoom control, though the static zoom level
    can be set as a parameter.
    """

    ptz = param.ClassSelector(PTZDevice,default=None,doc="""
       An instance of playerrobot.PTZDevice to be controlled.""")
    
    zoom = param.Number(default=120,bounds=(0,None),doc="""
       Desired FOV width in degrees.""")

    speed = param.Number(default=200,bounds=(0,None),doc="""
       Desired max pan/tilt speed in deg/sec.""")

    invert_amplitude = param.Boolean(default=False,doc="""
       Invert the sense of the amplitude signal, in order to get the
       appropriate ipsi-/contralateral sense of saccades.""")

    dest_ports = ["Saccade"]
    src_ports = ["State"]

    def start(self):
        pass 
    def input_event(self,conn,data):
        if conn.dest_port == "Saccade":
            # the data should be (amplitude,direction)
            amplitude,direction = data
            self.shift(amplitude,direction)

    def shift(self,amplitude,direction):

        self.debug("Executing shift, amplitude=%.2f, direction=%.2f"%(amplitude,direction))
        if self.invert_amplitude:
            amplitude *= -1

        # if the amplitude is negative, invert the direction, so up is still up.
        if amplitude < 0:
            direction *= -1
        angle = direction * pi/180

        pan,tilt,zoom = self.ptz.state_deg       
        pan += amplitude * cos(angle)
        tilt += amplitude * sin(angle)

        self.ptz.set_ws_deg(pan,tilt,self.zoom,self.speed,self.speed)
        ## self.ptz.cmd_queue.put_nowait(('set_ws_deg',
        ##                                (pan,tilt,self.zoom,self.speed,self.speed)))




    
        



"""
SOM-based learning functions for CFProjections.

$Id$
"""


from math import ceil

import param

from topo.base.arrayutil import L2norm, array_argmax
from topo.base.boundingregion import BoundingBox
from topo.base.cf import CFPLearningFn
from topo.base.patterngenerator import PatternGenerator
    
from topo.pattern import Gaussian


### JABHACKALERT: This class will be removed once the examples no
### longer rely upon it
class CFPLF_SOM(CFPLearningFn):
    """
    An abstract base class of learning functions for Self-Organizing Maps.

    This implementation is obsolete and will be removed soon.
    Please see examples/cfsom_or.ty for current SOM support.
    """
    __abstract = True

    learning_radius = param.Number(default=0.0,doc=
        """
        The radius of the neighborhood function to be used for
        learning.  Typically, this value will be set by the Sheet or
        Projection owning this CFPLearningFn, but it can also be set
        explicitly by the user.
        """)

    def __init__(self,**params):
        self.warning("CFPLF_SOM is deprecated -- see the example in cfsom_or.ty for how to build a SOM")
    
    def __call__(self, proj, input_activity, output_activity, learning_rate, **params):
        raise NotImplementedError



### JABHACKALERT: This class will be removed once the examples no
### longer rely upon it
class CFPLF_HebbianSOM(CFPLF_SOM):
    """
    Hebbian learning rule for CFProjections to Self-Organizing Maps.

    This implementation is obsolete and will be removed soon.
    Please see examples/cfsom_or.ty for current SOM support.
    """

    learning_radius = param.Number(default=0.0)
    
    crop_radius_multiplier = param.Number(default=3.0,doc=
        """
        Factor by which the radius should be multiplied,
        when deciding how far from the winner to keep updating the weights.
        """)
    
    neighborhood_kernel_generator = param.ClassSelector(PatternGenerator,
        default=Gaussian(x=0.0,y=0.0,aspect_ratio=1.0),
        doc="Neighborhood function")
    

    def __call__(self, iterator, input_activity, output_activity, learning_rate, **params):
        cfs = iterator.proj.cfs.tolist() # CEBALERT: convert to use flatcfs
        rows,cols = output_activity.shape

        # This learning function does not need to scale the learning
        # rate like some do, so it does not use constant_sum_connection_rate()
        single_connection_learning_rate = learning_rate
        
        ### JABALERT: The learning_radius is normally set by
        ### the learn() function of CFSOM, so it doesn't matter
        ### much that the value accepted here is in matrix and 
        ### not sheet coordinates.  It's confusing that anything
        ### would accept matrix coordinates, but the learning_fn
        ### doesn't have access to the sheet, so it can't easily
        ### convert from sheet coords.
        radius = self.learning_radius
        crop_radius = max(1.25,radius*self.crop_radius_multiplier)
        
        # find out the matrix coordinates of the winner
        #
        # NOTE: when there are multiple projections, it would be
        # slightly more efficient to calculate the winner coordinates
        # within the Sheet, e.g. by moving winner_coords() to CFSOM
        # and passing in the results here.  However, finding the
        # coordinates does not take much time, and requiring the
        # winner to be passed in would make it harder to mix and match
        # Projections and learning rules with different Sheets.
        wr,wc = array_argmax(output_activity)
        
        # Optimization: Calculate the bounding box around the winner
        # in which weights will be changed, to avoid considering those
        # units below.
        cmin = int(max(wc-crop_radius,0))
        cmax = int(min(wc+crop_radius+1,cols)) # at least 1 between cmin and cmax
        rmin = int(max(wr-crop_radius,0))
        rmax = int(min(wr+crop_radius+1,rows))

        # generate the neighborhood kernel matrix so that the values
        # can be read off easily using matrix coordinates.
        nk_generator = self.neighborhood_kernel_generator
        radius_int = int(ceil(crop_radius))
        rbound = radius_int + 0.5
        bb = BoundingBox(points=((-rbound,-rbound), (rbound,rbound)))

        # Print parameters designed to match fm2d's output
        #print "%d rad= %d std= %f alpha= %f" % (topo.sim._time, radius_int, radius, single_connection_learning_rate)

        neighborhood_matrix = nk_generator(bounds=bb,xdensity=1,ydensity=1,
                                           size=2*radius)

        for r in range(rmin,rmax):
            for c in range(cmin,cmax):
                cwc = c - wc 
                rwr = r - wr 
                lattice_dist = L2norm((cwc,rwr))
                if lattice_dist <= crop_radius:
                    cf = cfs[r][c]
                    rate = single_connection_learning_rate * neighborhood_matrix[rwr+radius_int,cwc+radius_int]
                    X = cf.get_input_matrix(input_activity)
                    cf.weights += rate * (X - cf.weights)

                    # CEBHACKALERT: see ConnectionField.__init__()
                    cf.weights *= cf.mask





"""
Panels for displaying tuning curves.

Uses a PlotGroup to generate the pylab plots 
which are currently displayed separately from the gui.

$Id$
"""

import topo

from topo.base.projection import ProjectionSheet

from plotgrouppanel import PlotGroupPanel


## CEBALERT: same as for featurecurveplotgroup: shares code with templateplotgrouppanel
# Should be changed to inherit from UnitPGPanel, or whatever is created to handle the PlotGroup
# hierarchy.
class FeatureCurvePanel(PlotGroupPanel):

    # CEBHACKALERT: to which types of sheet is this plotgroup supposed to be applicable?
    sheet_type = ProjectionSheet

    @classmethod
    def valid_context(cls):
        if topo.sim.objects(cls.sheet_type).items():
            return True
        else:
            return False


    def __init__(self,master,plotgroup,**params):
	PlotGroupPanel.__init__(self,master,plotgroup,**params)

        self.pack_param("sheet",parent=self.control_frame_3,
            on_modify=self.sheet_change,side='left',expand=1,
            widget_options={'new_default':True,
                            'sort_fn_args':{'cmp':lambda x, y: cmp(-x.precedence,-y.precedence)}})

        self.pack_param("x",parent=self.control_frame_4)
        self.pack_param("y",parent=self.control_frame_4)

        # remove currently irrelevant widgets (plots are drawn in a separate window by pylab)
        # CEBNOTE: when plots are in this window, remove this line.
        for name in ['Enlarge','Reduce','Back','Fwd']: self.hide_param(name)

        self.auto_refresh= False
        if self.plotgroup.plot_immediately:
            self.refresh()

        self.sheet_change()


        
    def sheet_change(self):
        s = self.sheet
        l,b,r,t = s.bounds.lbrt()

        x = self.get_parameter_object('x')
        y = self.get_parameter_object('y')

        x.bounds=(l,r)
        y.bounds=(b,t)

        self.x = 0.0
        self.y = 0.0

        if 'x' and 'y' in self.representations:
            w1,w2=self.representations['x']['widget'],self.representations['y']['widget']
            w1.set_bounds(*x.bounds)
            w2.set_bounds(*y.bounds)
            
            w1.tag_set();w2.tag_set()



    def setup_plotgroup(self):
        super(FeatureCurvePanel,self).setup_plotgroup()
        self.populate_sheet_param()


    def populate_sheet_param(self):
        sheets = topo.sim.objects(self.sheet_type).values() 
        self.plotgroup.params()['sheet'].objects = sheets
        self.plotgroup.sheet = sheets[0] # CB: necessary?

    def _plot_title(self):
        return self.plotgroup.name+' at time ' + topo.sim.timestr(self.plotgroup.time)


    def display_labels(self):
        """Plots are displayed in new windows, so do not add any labels."""
        pass










"""
The Test Pattern window allows input patterns to be previewed.

$Id$
"""

## Notes:
# * need to remove audigen pgs
# * missing disparity flip hack (though see JABHACKALERT below)
# * values like pi are written over
# * need to sort the list of Pattern generators

## Needs to be upgraded to behave how we want:
### JABHACKALERT: Should use PatternPresenter (from
### topo.command.analysis), which will allow flexible support for
### making objects with different parameters in the different eyes,
### e.g. to test ocular dominance or disparity.

# CBENHANCEMENT: add 'use for learning' to install current pattern
# (saving previous ones)?

from Tkinter import Frame

import param
import paramtk as tk

import topo

from topo.base.functionfamily import PatternDrivenAnalysis
from topo.base.sheetview import SheetView
from topo.base.patterngenerator import PatternGenerator, Constant
from topo.misc.generatorsheet import GeneratorSheet
from topo.command import pattern_present
from topo.plotting.plot import make_template_plot
from topo.plotting.plotgroup import SheetPlotGroup

from plotgrouppanel import SheetPanel


# CEBALERT: this uses make_template_plot(), so how is it not a
# TemplatePlotGroup? 
class TestPatternPlotGroup(SheetPlotGroup):

    def sheets(self):
        if hasattr(self,'_sheets'):
            return self._sheets
        else:
            return super(TestPatternPlotGroup,self).sheets()

    def _generate_plots(self):
        dynamic_plots = []
        for kw in [dict(sheet=sheet) for sheet in self.sheets()]:
            sheet = kw['sheet']
            new_view = SheetView((sheet.input_generator(),sheet.bounds),
                                  sheet.name,sheet.precedence,topo.sim.time())
            sheet.sheet_views['Activity']=new_view
            channels = {'Strength':'Activity','Hue':None,'Confidence':None}
        
            ### JCALERT! it is not good to have to pass '' here... maybe a test in plot would be better
            dynamic_plots.append(make_template_plot(channels,sheet.sheet_views,
                                                    sheet.xdensity,sheet.bounds,self.normalize,
                                                    name=''))

        return self._static_plots[:]+dynamic_plots



class TestPattern(SheetPanel,PatternDrivenAnalysis):

    sheet_type = GeneratorSheet
    
    dock = param.Boolean(False)

    edit_sheet = param.ObjectSelector(doc="""
        Sheet for which to edit pattern properties.""")

    plastic = param.Boolean(default=False,doc="""
        Whether to enable plasticity during presentation.""")
    
    duration = param.Number(default=1.0, softbounds=(0.0,10.0),doc="""
        How long to run the simulation for each presentation.""")

    Present = tk.Button(doc="""Present this pattern to the simulation.""")

    pattern_generator = param.ClassSelector(default=Constant(), class_=PatternGenerator, doc="""
        Type of pattern to present. Each type has various parameters that can be changed.""")



    def __init__(self,master,plotgroup=None,**params):
        plotgroup = plotgroup or TestPatternPlotGroup()
        
        super(TestPattern,self).__init__(master,plotgroup,**params)
        
        self.auto_refresh = True

        self.plotcommand_frame.pack_forget()
        for name in ['pre_plot_hooks','plot_hooks','Fwd','Back']:
            self.hide_param(name)

        edit_sheet_param = self.get_parameter_object('edit_sheet')
        edit_sheet_param.objects = self.plotgroup.sheets()

        self.pg_control_pane = Frame(self) #,bd=1,relief="sunken")
        self.pg_control_pane.pack(side="top",expand='yes',fill='x')
        
        self.params_frame = tk.ParametersFrame(
            self.pg_control_pane,
            parameterized_object=self.pattern_generator,
            on_modify=self.conditional_refresh,
            msg_handler=master.status)

        self.params_frame.hide_param('Close')
        self.params_frame.hide_param('Refresh')

        # CEB: 'new_default=True' is temporary so that the current
        # behavior is the same as before; shoudl make None the
        # default and mean 'apply to all sheets'.
        self.pack_param('edit_sheet',parent=self.pg_control_pane,on_modify=self.switch_sheet,widget_options={'new_default':True,'sort_fn_args':{'cmp':lambda x, y: cmp(-x.precedence,-y.precedence)}})
        self.pack_param('pattern_generator',parent=self.pg_control_pane,
                        on_modify=self.change_pattern_generator,side="top")
        
        present_frame = Frame(self)
        present_frame.pack(side='bottom')

        self.pack_param('plastic',side='bottom',parent=present_frame)
        self.params_frame.pack(side='bottom',expand='yes',fill='x')
        self.pack_param('duration',parent=present_frame,side='left')
        self.pack_param('Present',parent=present_frame,on_set=self.present_pattern,side="right")


    def setup_plotgroup(self):
        super(TestPattern,self).setup_plotgroup()
        
        # CB: could copy the sheets instead (deleting connections etc)
        self.plotgroup._sheets = [GeneratorSheet(name=gs.name,
                                                 nominal_bounds=gs.nominal_bounds,
                                                 nominal_density=gs.nominal_density)
                                  for gs in topo.sim.objects(GeneratorSheet).values()]
        self.plotgroup._set_name("Test Pattern")


    def switch_sheet(self):
        if self.edit_sheet is not None:
            self.pattern_generator = self.edit_sheet.input_generator
        self.change_pattern_generator()

        
    def change_pattern_generator(self):
        """
        Set the current PatternGenerator to the one selected and get the
        ParametersFrameWithApply to draw the relevant widgets
        """
        # CEBALERT: if pattern generator is set to None, there will be
        # an error. Need to handle None in the appropriate place
        # (presumably tk.py).
        self.params_frame.set_PO(self.pattern_generator)

        for sheet in self.plotgroup.sheets():
            if sheet==self.edit_sheet:
                sheet.set_input_generator(self.pattern_generator)
        
        self.conditional_refresh()


    def refresh(self):
        """
        Simply update the plots: skip all handling of history.
        """
        self.refresh_plots()

        
    def present_pattern(self):
        """
        Move the user created patterns into the GeneratorSheets, run for
        the specified length of time, then restore the original
        patterns.
        """
        topo.sim.run(0.0)  # ensure EPs are start()ed
        
        topo.sim.state_push()
        for f in self.pre_presentation_hooks: f()
        input_dict = dict([(sheet.name,sheet.input_generator) \
                           for sheet in self.plotgroup.sheets()])
        pattern_present(input_dict,self.duration,
                        plastic=self.plastic,overwrite_previous=False)
        topo.guimain.auto_refresh()
        for f in self.post_presentation_hooks: f()
        topo.sim.state_pop()
        
